#open "item";;
#open "microq";;

let raiseAndPrint x = ((match x with Failure y -> print_string (y) | _ -> print_string ("urgh")); print_newline (); raise (x));; 

let float_of_item x = match x with INum y -> y | _ -> raiseAndPrint (Failure "runtime error, expected num");;
let fuzz_of_item x = match x with IFuzz y -> y | _ -> raiseAndPrint (Failure "runtime error, expected fuzz");;
let string_of_item x = match x with IString y -> y | _ -> raiseAndPrint (Failure "runtime error, expected string");;

let print x = print_string x; flush stdout;;
let print_err x = prerr_string x; flush stderr;;

let suffix s n = sub_string s n (string_length s - n);;

let rec split d s = 
  try let delim = index_char s d in
    (sub_string s 0 delim :: split d (suffix s (delim+1)))
  with Not_found -> [s];;

let rec join sd xs = 
  match xs with 
      [] -> ""
    | [x] -> x
    | h::tl -> h ^ sd ^ join sd tl;;

let rec item_to_toks x = match x with 
    (INum n) -> [ "N";  (string_of_float n) ]
  | (IPair (a, b)) -> [ "P" ] @ (item_to_toks a) @ (item_to_toks b) 
  | (IInl a) -> [ "L" ] @ item_to_toks a
  | (IInr a) -> [ "R" ] @ item_to_toks a
  | IUnit -> ["U"]
  | IString s -> ["S" ; s]
  | IBool b -> if b then ["T"] else ["F"]
  | IBag _ -> failwith "serializing bag?"
  | IFn _ -> failwith "serializing fn?"
  | IFuzz _ -> failwith "serializing fuzz?"  (* XXX *)
  | NULL -> ["X"];;

let rec parse_toks x = match x with 
    "N" :: n :: tl  -> (INum (float_of_string n) , tl)
  | "P" :: tl -> 
      let (a, tl) = parse_toks tl in
      let (b, tl) = parse_toks tl in (IPair(a, b), tl)
  | "U" :: tl -> (IUnit, tl)
  | "L" :: tl -> let (a, tl) = parse_toks tl in (IInl a, tl)
  | "R" :: tl -> let (a, tl) = parse_toks tl in (IInr a, tl)
  | "S" :: s :: tl -> (IString s, tl)
  | "T" :: tl -> (IBool true, tl)
  | "F" :: tl -> (IBool false, tl)
  | "X" :: tl -> (NULL, tl)
  | _ -> raiseAndPrint (Failure ("parse error during deserialization: [" ^ (hd x) ^ "]"));;

let toks_to_item x = match parse_toks x with (a, []) -> a 
  | _ -> raiseAndPrint (Failure "extra tokens during deserialization");;

let delim = char_of_int 1;;
let string_to_item x = toks_to_item (split delim x);;
let item_to_string x = join (string_of_char delim) (item_to_toks x);;

let itrue x = x = IBool true;;
let iequal x y = IBool (x = y) ;;
let igt x y = IBool (x > y) ;;
let ilt x y = IBool (x < y) ;;
let iapp x y = match x with IFn f -> 
    (match y with 
     NULL -> NULL
     | _ -> f y)
| _ -> raiseAndPrint (Failure "runtime type error: applying a nonfunction");;
let iconj x y = IBool (match (x, y) with (IBool true, IBool true) -> true | _ -> false);;
let idisj x y = IBool (match (x, y) with (IBool false, IBool false) -> false | _ -> true);;
let iplus x y = match (x, y) with
    (INum xx, INum yy) -> INum (xx +. yy)
  | (IString xx, IString yy) -> IString (xx ^ yy)
  | _ -> raiseAndPrint (Failure "runtime type error: adding nonnumbers");;
let iminus x y = match (x, y) with
    (INum xx, INum yy) -> INum (xx -. yy)
  | _ -> raiseAndPrint (Failure "runtime type error: subtracting nonnumbers");;
let idiv x y = match (x, y) with
    (INum xx, INum yy) -> INum (xx /. yy)
  | _ -> raiseAndPrint (Failure "runtime type error: dividing nonnumbers");;
let imult x y = match (x, y) with
    (INum xx, INum yy) -> INum (xx *. yy)
  | _ -> raiseAndPrint (Failure "runtime type error: dividing nonnumbers");;


let dethunk x = match x with IFuzz f -> f () | _ -> raiseAndPrint (Failure "runtime type error: dethunking a nonthunk");;
let sample k f = k (dethunk f);;
let return = IFn (fun x -> IFuzz (fun () -> x));;

let lap sigma = 
  let s = 2.0 *. float_of_int (random__int 2) -. 1.0 in
    s *. sigma *. log(random__float 1.0);;

let fuzz  = IFn (fun x -> IFuzz (fun () -> iplus x (INum (lap 1.0))));; 
(*let bagsize = IFn (fun x -> match x with IBag xs -> INum (float_of_int (list_length xs)) | _ -> raiseAndPrint (Failure "runtime type error: bagsizing a nonbag"));;*)

let clip = IFn (fun x -> match x with INum x -> 
                  INum (if x >. 1.0 then 1.0 else if x <. 0.0 then 0.0 else x)  
                  | _ ->  raiseAndPrint (Failure "runtime type error: clipping a nonnumber"));;

(* Microquery inner loop *)

(* let int_of_string = (fun x -> int_of_float (float_of_string x));; *)

(*let mq f timeout =
    (fun x -> let row = string_to_item x in
		  item_to_string (iapp f row)
    );;
*)
let mq f timeout numDBexpected defaultVal =
  (fun x -> (
     let _ = pre_mqs x numDBexpected defaultVal in 
     let r = checkpoint timeout in
     match (nth_char r 0) with
       `1` -> (IBag (int_of_string(sub_string r 1 ((string_length r)-1))))
     | `2` -> (IPair ((IBag (int_of_string(sub_string r 1 ((index_char r `,`)-1)))),
                      (IBag (int_of_string(sub_string r ((index_char r `,`)+1) ((string_length r)-(index_char r `,`)-1))))
              ))
     |  _  -> let row = string_to_item r in
	        let result = item_to_string (iapp f row) in
		   let _ = store_result(result) in 
                     (IBag 0) (* Just to typecheck, will never be run*)
  ));;

(* bagmap : ('a -> 'b) -> 'a -> Num -> 'a bag -> 'b bag  *)
let bagmap = IFn (fun f -> IFn (fun default -> IFn (fun t -> IFn (fun bag ->
  let timeout = (int_of_float (float_of_item t)) in
    let xs = match bag with 
      IBag xs -> xs 
    | _       -> raiseAndPrint (Failure "runtime error: bagmapping nonbag") in
         mq f timeout 1 (item_to_string default) xs))));;

(* bagsplit : ('a -> bool) -> Num -> 'a bag -> 'a bag * 'a bag *)
let bagsplit = IFn (fun f -> IFn (fun t -> IFn (fun bag ->
  let timeout = (int_of_float (float_of_item t)) in
    let xs = match bag with 
      IBag xs -> xs 
    | _       -> raiseAndPrint (Failure "runtime error: bagmapping nonbag") in
         mq f timeout 2 "F;" xs)));;

let bagsize = IFn (fun x -> match x with IBag xs -> INum (cbagsize xs) | _ -> raiseAndPrint (Failure "runtime type error: bagsizing a nonbag"));;


let bagsum = IFn (fun x -> match x with IBag xs -> INum (cbagsum xs) | _ -> raiseAndPrint (Failure "runtime type error: bagsumming a nonbag"));;


let num2string 
    = IFn (fun x -> IString (match x with (INum xx) -> string_of_float xx | _ -> raiseAndPrint (Failure "runtime type error: num2string on nonnumber")));;

let decons_pair x = match x with IPair p -> p | _ -> raiseAndPrint (Failure ("runtime type error, deconstructing as pair: " ^ item_to_string x));;
