#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/mman.h>
#include <string.h>
#include <limits.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <sched.h>
#include <setjmp.h>
// caml includes
#include <config.h>
#include <alloc.h>
#include <memory.h>
#include <misc.h>
#include <mlvalues.h>
#include "tracer.h"
#include <assert.h>

#define panic(a...) do { fprintf(stderr, a); fprintf(stderr, "\n"); dump_trace_buffer(); if(child_pid!=0) {kill(child_pid, SIGTERM);} exit(1);} while (0)
#define dprintf(a...) do { if (clFlags & FLAG_VERBOSE) { fprintf(stderr, a); fprintf(stderr, "\n"); } } while (0)

#define FUDGE_OFFSET 1000ULL
#define FUDGE_FACTOR 10ULL
#define CYCLES_PER_USEC 3060ULL
#define FUDGE_START 1000ULL
#define FUDGE_END 10000UL
#define SAVE_LEN 9 // in words
#define RETURN_VALUE_SIZE 1024
#define FLAG_FUDGE 1
#define FLAG_NOTIMEOUT 2
#define FLAG_VERBOSE 4
#define FLAG_PROFILE 8
#define MAX_NUM_DBS (100)


#define PROFILER

// *** EXTERNS **************************************************************

extern void minor_collection();
extern char *young_start, *young_ptr, *young_end;
extern char *heap_start, *heap_end;
extern int gc_disabled; /* defined in minor_gc.c in modified ocaml runtime */
extern int force_long_term;

/* List of extern values from the .save segment */
extern value *c_roots_head;
extern value * arg_stack_low;
extern value * ret_stack_low;
extern value * extern_asp;
extern value * extern_rsp;
extern struct trap_frame * tp;

extern void (*alarmHandler)();

// *** GLOBAL ***************************************************************

unsigned int clFlags = 0;
int haveRoot = 1;
int grandTotalTimeouts = 0;

// *** CHECKPOINT ***********************************************************

value *c_roots_head__;
value * arg_stack_low__;
value * ret_stack_low__;
value * extern_asp__;
value * extern_rsp__;
char * young_ptr__;
struct trap_frame * tp__;
int timeoutBackup = 0;
extern sigjmp_buf jmpbuf;

// *** TIMER ****************************************************************

volatile unsigned long long wakeupTime;
volatile unsigned long long timePlanEnd = 0;
volatile int timerready;
int child_pid;
int interpreter_pid;
int timeout = 0;
unsigned long long timeStartC = 0;
unsigned long long timeStartMQ = 0;
unsigned long long burnBasis;
unsigned long long maxRecoveryTime;

// *** MICROQUERY ***********************************************************

char microqReturnValue[RETURN_VALUE_SIZE];
char microqDefaultValue[RETURN_VALUE_SIZE];
int microqReturnLen;
int totalMQs;
int isWarmup;

#ifdef PROFILER
unsigned long long mqMax = 0;
int totalTimeouts = 0;
int totalFinished = 0;
#endif

// *** DATABASE *************************************************************

typedef char *dbHandle;
dbHandle database[MAX_NUM_DBS];
int numDBs = 0;
dbHandle currentDBhandle;
int currentNumDBexpected;
dbHandle db1, db2;
int dbi1, dbi2;
int __numRows;
int __numBytesPerRow;

void print_usage()
{
  fprintf(stderr, "Usage: runquery [--no-root] <preprocDB>\n");
  exit(1);
}

int alloc_database()
{
  assert(numDBs<MAX_NUM_DBS);
  
  database[numDBs] = (char*)malloc(__numRows * __numBytesPerRow);
  if (!database[numDBs])
    panic("Out of memory in alloc_database");
    
  dprintf("new database allocated, handle=%d", numDBs);
    
  return numDBs++;
}

value parse_arguments(value args)
{
  char buf[2000];
  strncpy(buf, (char*)args, sizeof(buf));
  
  char *argv[20];
  argv[0] = strtok(buf, " ");
  for (int i=1; i<20; i++)
    argv[i] = argv[i-1] ? strtok(NULL, " ") : NULL;
    
  char *dbname = NULL;
    
  for (int i=1; i<20; i++) {
    if (argv[i]) {
      if (!strcmp(argv[i], "--no-root")) {
        haveRoot = 0;
      } else if (!strcmp(argv[i], "--verbose")) {
        clFlags |= FLAG_VERBOSE;
      } else if (!strcmp(argv[i], "--profile")) {
        clFlags |= FLAG_PROFILE;
      } else if (!strcmp(argv[i], "--mode")) {
        char *profile = argv[++i];
        
        if(!strcmp((char *)profile, "fudge")) {
          clFlags |= FLAG_FUDGE;
        } else if(!strcmp((char *)profile, "notimeout")) {
          clFlags |= FLAG_NOTIMEOUT;
        } else if(!strcmp((char *)profile, "run")) {
          /* nothing */
        } else {
          panic( "profile was [%s], should be one of [fudge|notimeout|run]", profile);
        }

        if(clFlags & FLAG_FUDGE) {
          fprintf(stderr, "program profiling fudge\n");
        }
        if(clFlags & FLAG_NOTIMEOUT) {
          fprintf(stderr, "program profiling computation time without timeouts\n");
        }
      } else {
        if (!dbname)
          dbname = argv[i];
        else 
          print_usage();
      }
    }
  }
  
  if (!dbname)
    print_usage();
    
  int h = open(dbname, O_RDONLY);
  if (h<0)
    panic("Cannot read database: %s", dbname);
    
  int r;
  r = read(h, &__numRows, sizeof(__numRows));
  if (r<sizeof(__numRows))
    panic("Cannot read: %s", dbname);
  r = read(h, &__numBytesPerRow, sizeof(__numBytesPerRow));
  if (r<sizeof(__numBytesPerRow))
    panic("Cannot read: %s", dbname);

  assert(__numBytesPerRow <= RETURN_VALUE_SIZE);
    
  dprintf("Database %s has %d rows of %d bytes", dbname, __numRows, __numBytesPerRow);

  int dbi = alloc_database();
  dbHandle db = database[dbi];
  if (!db)
    panic("Out of memory -- database too large?");
    
  r = read(h, db, __numRows * __numBytesPerRow);
  if (r < __numRows * __numBytesPerRow)
    panic("Cannot read the database (only %d bytes read)", r);

  return Val_int((int)dbi);
}


/* checkpoint saves state before a new microquery. It saves
   interpreter state in some backup variables. the interpreter
   intercepts calls to checkpoint and setjmps just before entering the
   function. longjump will thus jump back into the interpreter just
   before checkpoint, so checkpoint will run a second time, but with
   -1 as timeoutInMilliseconds.

   checkpoint returns: 

   - 0, if we have successfully saved state and a set ualarm.
   
   - 1, if we just returned from a longjmp after we timed out.
 */

void handleAlarm() {
  stop_bucket(3);
  stop_bucket(4);
  stop_bucket(5);
  start_bucket(5, "microquery in-between");
  add_event("ALRM");
  //wakeupTime = 0;
  siglongjmp(jmpbuf, 1);
}

value printProfile() {
  fprintf(stderr, "Max time taken was %lld cycles, %f us, there were %d timeouts, %d nontimeouts, set timeout was %d\n", mqMax, (mqMax / 3060.0), totalTimeouts, totalFinished, timeoutBackup);
  return Val_unit;
}

value pre_mqs(value dbHandleV, value numDBexpectedV, value defaultValue) {
  
  start_bucket(2, "macroqueries");
  start_bucket(5, "microquery in-between");
  burnBasis = rdtsc();
  
  if (strlen((char*)defaultValue) > (__numBytesPerRow-1))
    panic("Default value is too big (but we haven't looked at the database yet)");
    
  int index = Int_val(dbHandleV);
  assert((0<=index) && (index<numDBs));
  currentDBhandle = database[index];
  currentNumDBexpected = Int_val(numDBexpectedV);
  
  dprintf("pre_mqs(dbHandle=%d, numDBexpected=%d, default=[%s]", index, currentNumDBexpected, (char *) defaultValue);
  
  dbi1 = alloc_database();
  dprintf("DB1=%d", dbi1);
  db1 = database[dbi1];
  if (currentNumDBexpected == 2) {
    dbi2 = alloc_database();
    dprintf("DB2=%d", dbi2);
    db2 = database[dbi2];
  } else {
    dbi2 = -1;
    db2 = NULL;
  }
  
  totalMQs = 0;

  mqMax = 0;
  totalTimeouts = 0;
  totalFinished = 0;
  maxRecoveryTime = 0;
  
  memset(microqDefaultValue, 0, RETURN_VALUE_SIZE);
  memcpy(microqDefaultValue, (char *) defaultValue, RETURN_VALUE_SIZE);
  minor_collection();
  assert(young_ptr == young_start);
  force_long_term = 1;
  dprintf("pre_mqs ends");
  return Val_unit;
}

void copyResult() {
  assert((totalMQs>=0) && (totalMQs < __numRows));
  if (isWarmup)
    return;

  if (currentNumDBexpected == 1) {
    
    /* We are doing a bagmap */
    
    if (microqReturnLen > 0) {
      totalFinished++;
      strcpy(db1 + (totalMQs * __numBytesPerRow), microqReturnValue);
    } else {
      totalTimeouts ++;
      grandTotalTimeouts ++;
      strcpy(db1 + (totalMQs * __numBytesPerRow), microqDefaultValue);
    }
  } else {
  
    /* We are doing a bagsplit */

    if (microqReturnLen > 0) {
      totalFinished++;
      if (microqReturnValue[0] == 'T') {
        strcpy(db1 + (totalMQs * __numBytesPerRow), currentDBhandle + (totalMQs * __numBytesPerRow));
        strcpy(db2 + (totalMQs * __numBytesPerRow), "X"); /* the null value */
      } else {
        strcpy(db1 + (totalMQs * __numBytesPerRow), "X"); /* the null value */
        strcpy(db2 + (totalMQs * __numBytesPerRow), currentDBhandle + (totalMQs * __numBytesPerRow));
      }
    } else {
      totalTimeouts ++;
      grandTotalTimeouts ++;
      strcpy(db1 + (totalMQs * __numBytesPerRow), "X"); /* the null value */
      strcpy(db2 + (totalMQs * __numBytesPerRow), currentDBhandle + (totalMQs * __numBytesPerRow));
    }
  }
}

void fudgeWait()
{
  unsigned long long arrivalTime = rdtsc();
  unsigned long long timeEndFudge = burnBasis + ((1+__numRows)*(timeout+FUDGE_FACTOR) + FUDGE_OFFSET) * CYCLES_PER_USEC;
  
  if(clFlags & FLAG_FUDGE) {
    fprintf(stderr, "In fudge mode, the average amount of cleanup per row is %lld cycles, %lld us\n", (arrivalTime - burnBasis)/__numRows, ((arrivalTime - burnBasis)/(CYCLES_PER_USEC *__numRows)));
  }
  
  if ((!(clFlags & (FLAG_FUDGE|FLAG_NOTIMEOUT))) && (arrivalTime > timeEndFudge)) 
    panic("We arrived into the wait %lld cycles late\n", arrivalTime - timeEndFudge);

  /* Benchmarking for recovery time (used in --mode fudge) */

  unsigned long long recoveryTime = arrivalTime - timePlanEnd;
  if(recoveryTime > maxRecoveryTime) {
    maxRecoveryTime = recoveryTime;
  }
  
  /* Burn the remaining cycles */
  
  start_bucket(6, "microquery fudge wait");
  if((!(clFlags & (FLAG_FUDGE|FLAG_NOTIMEOUT)))) 
    while(rdtsc() < timeEndFudge);
  stop_bucket(6);
    
  add_event("wait");  
}

value checkpoint(value timeoutInMilliseconds) {

  timeout = Int_val(timeoutInMilliseconds);
  
  dprintf("checkpoint(%d)", timeout);
  
  if(timeout>=0) {
    timeoutBackup = timeout;
    
    timeStartC = rdtsc();
    c_roots_head__ = c_roots_head;
    arg_stack_low__  = arg_stack_low;
    ret_stack_low__  = ret_stack_low;
    extern_asp__     = extern_asp;
    extern_rsp__     = extern_rsp;
    tp__             = tp;
    young_ptr__      = young_ptr;
    
    // TODO Bound the size that the stack can grow.
    
    alarmHandler = handleAlarm;	
    isWarmup = 1;
  }
  else {
    add_event("ljmp");
    // Restore everything from checkpoint
    c_roots_head    = c_roots_head__;
    arg_stack_low   = arg_stack_low__;
    ret_stack_low   = ret_stack_low__;
    extern_asp      = extern_asp__;
    extern_rsp      = extern_rsp__;
    tp              = tp__;
    young_ptr       = young_ptr__;
    timeout         = timeoutBackup;    
    copyResult();
    if (isWarmup) {
      isWarmup = 0;
    } else {
      totalMQs ++;
    }
  }

  /* Now we decide what to return */

  dprintf(" - totalMQs = %d/%d", totalMQs, __numRows);
  if (totalMQs == __numRows) {
  
    /* We're done */
    
    if(clFlags & FLAG_FUDGE)
      fprintf(stderr, "Maximum microquery recovery time was %lld cycles\n", maxRecoveryTime);

    force_long_term = 0;
    gc_disabled = 0;
    
    if (clFlags & (FLAG_FUDGE|FLAG_PROFILE))
      printProfile();
    
    char buf[2000];
    if (currentNumDBexpected == 2)
      sprintf(buf, "2%d,%d", dbi1, dbi2);
    else
      sprintf(buf, "1%d", dbi1);
      
    dprintf("checkpoint returning %s", buf);
    stop_bucket(5);
    fudgeWait();
    stop_bucket(2);
      
    return copy_string(buf);
  } else {
    assert(young_ptr == young_start);

    microqReturnLen = 0;
    gc_disabled    = 1;
    force_long_term= 0;

    stop_bucket(5);
    
    timeStartMQ = rdtsc();
    
    if(clFlags & FLAG_FUDGE) {
      timePlanEnd = 1;
    } else {
      timePlanEnd = (CYCLES_PER_USEC * timeout)+ timeStartMQ;
    }

    //fprintf(stderr, "wakeupTime: %lld is %lld cycles from now\n", wakeupTime, (wakeupTime - timeStartC));
    dprintf(" - running f in ML world");
    char * thestring = currentDBhandle + (totalMQs*__numBytesPerRow);
    dprintf(" - string at %p, db handle %p, %d %d", thestring, currentDBhandle, totalMQs, __numBytesPerRow);
    dprintf(" - row is [%s]", thestring);
    add_event2("micq", totalMQs);
    
    start_bucket(3, "microquery computation");
    if(!(clFlags & FLAG_NOTIMEOUT)) {
      wakeupTime = timePlanEnd;
    }
    return copy_string(thestring);
  }
}

value store_result(value result) {
  dprintf(" - storing result");
  unsigned long long mqTime = (rdtsc() - timeStartMQ);
  if (!isWarmup && (mqTime > mqMax))
    mqMax = mqTime;
    
  add_event("comp");
  
  int len = strlen((char*)result);
  assert(len>0);
  
  if (len < __numBytesPerRow) {
    dprintf(" - result: [%s]", (char*)result);
    strcpy(microqReturnValue, (char *) result);
    microqReturnLen = len;
  } else {
    dprintf(" - result is too big");
    /* It's too big to fit into a row of the database, so we time out */
  }
  
  if(clFlags & FLAG_NOTIMEOUT) {
    stop_bucket(3);
    start_bucket(5, "microquery in-between");
    longjmp(jmpbuf, 1);
  }
  
  dprintf("end of store_result");
  stop_bucket(3);
  start_bucket(4, "microquery waiting");
  while(1);
  return Val_unit; // Just to type check. This function never returns.
}

value stub_time (value t) {
  return Val_int(time(NULL));
}

int timerFunction(void * nothing) {
  
  cpu_set_t x;
  CPU_ZERO(&x);
  CPU_SET(0, &x);
  sched_setaffinity( getpid(), sizeof(cpu_set_t), &x);
  timerready = 1;
  while(1) {
    if(wakeupTime > 0) {
      while(rdtsc() < wakeupTime) {}
      wakeupTime = 0;
      kill(interpreter_pid, SIGALRM);
    }
  }
}


value startProgram() {
  signal(SIGALRM, handleAlarm);
  
  dprintf("start program");
  
  cpu_set_t x;
  CPU_ZERO(&x);
  CPU_SET(1, &x);
  sched_setaffinity( getpid(), sizeof(cpu_set_t), &x);

  // Set up a new thread for timing.
  
  char * timerStackBottom = (char *) malloc(4096);
  interpreter_pid = getpid();
  timerready = 0;
  
  if (!(clFlags & FLAG_NOTIMEOUT)) {
    dprintf("cloning thread");
    child_pid = clone(timerFunction, timerStackBottom + 4088, 
  		    CLONE_VM | CLONE_FILES, 
	  	    NULL);
    while(!timerready);
  
    assert(child_pid>0);
  }
  
  add_event("btch");
  for (char *ptr = young_start; ptr < young_end; ptr += 4096) {
    char x = *((volatile char *)ptr);
    *((volatile char *)ptr) = x;
  }
  
  if (haveRoot) {
    struct rlimit rlim;
    getrlimit(RLIMIT_MEMLOCK, &rlim);
    getrlimit(RLIMIT_DATA, &rlim);

    int ret = mlockall(MCL_FUTURE);
    if (ret != 0)
      panic("Locking fails (%s)", strerror(errno));

    struct sched_param prio;
    prio.sched_priority = 99;
    if(sched_setscheduler(0, SCHED_FIFO, &prio)) 
      panic("Failed to set real-time priority\n");
  } else {
    fprintf(stderr, "Using --no-root mode - THIS IS NOT SAFE!!!\n");
  }

  add_event("tchp");
  start_bucket(1, "Fuzz program");
  
  return Val_unit;
}

value endProgram() {
  stop_bucket(1);
  if (!(clFlags & FLAG_NOTIMEOUT)) 
    kill(child_pid, SIGTERM);
  fprintf(stdout, "%d timeout(s) total\n", grandTotalTimeouts);
}

value cbagsize(value dbhandleV)
{
  assert((0<=Int_val(dbhandleV)) && (Int_val(dbhandleV)<numDBs));
  dbHandle db = database[Int_val(dbhandleV)];
  dprintf("cbagsize(%d)", Int_val(dbhandleV));
  
  int i, total=0;
  for (i=0; i<__numRows; i++) {
    if (*(db + (__numBytesPerRow*i)) != 'X')
      total ++;
  }
  
  return copy_double(total);
}

value cbagsum(value dbhandleV)
{
  assert((0<=Int_val(dbhandleV)) && (Int_val(dbhandleV)<numDBs));
  dbHandle db = database[Int_val(dbhandleV)];
  dprintf("cbagsum(%d)", Int_val(dbhandleV));
  dprintf("=== %d", Int_val(dbhandleV));  
  
  double d = 0;
  int i;
  for (i=0; i<__numRows; i++) {
    char *theRow = db + (__numBytesPerRow*i);
    dprintf("%d: %s (s=%.2f)\n", i, theRow, d);
    
    assert((theRow[0] == 'N') || (theRow[0] == 'X'));
    if (theRow[0] == 'N') /* don't forget the 0x01 */
      d += atof(&theRow[2]);
  }
  
  return copy_double(d);
}
