#ifndef __il_h__
#define __il_h__

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "typechecker.h"
#include "context.h"

/*
 * This header file defines the classes used in the 'intermediate language', which is used for type checking
 * and for evaluation. This is different from the abstract syntax tree (AST), which is generated by the 
 * parser and must be translated into AST first, using the TranslationVisitor.
*/

#define SAME_SENSITIVITY(a,b) ((isinf(a) && isinf(b)) || (((a)<(b)+0.001) && ((a)>(b)-0.001)))

class ASTVisitor;

/* === BASIC CLASSES ============================================================================================= */

/* ASTBase: The base class of all nodes in an AST tree that can be connected to something in the original source code
  (expressions and types). The SourceContext contains the file name, line number, position, etc., which we
  may want to display if there is an error. */

class ASTBase 
{
protected:
  SourceContext *context;

public:
  ASTBase(SourceContext *context)
    { this->context = context; };

  SourceContext *getContext() { return context; };    
  virtual void accept(ASTVisitor *worker) = 0;
};

/* Sometimes we need to find out whether the root of a subtree (usually known as a ASTType*) is of a specific type, 
   so that it can be safely cast. This is done by calling getTypeCode() on the root of the subtree, which returns
   one of the following values. */

enum typeCode { TYP_BASE, TYP_VAR, TYP_UNION, TYP_TENSOR, TYP_NAMED, TYP_AMPERSAND, TYP_FUZZY, TYP_UNIT, TYP_LAMBDA, TYP_MU, TYP_SET };

/* ASTType: The base class of all type nodes. Types may be cloned, tested for equality, and resolved
   to other types (in the case of a typedef). When a mu type is used, it may also be necessary to bind type variables
   to a specific type. */

class ASTType : public ASTBase
{
public:
  ASTType(SourceContext *context) : ASTBase(context) {};
  virtual ASTType *clone() = 0;
  virtual bool equals(ASTType *peer, TCContext *ctx, ASTType **diffLocal = NULL, ASTType **diffPeer = NULL);
  static bool unify(ASTType **typeA, ASTType **typeB, TCContext *ctx, ASTType **ctxExt = NULL);
  virtual typeCode getTypeCode() = 0;
  virtual bool needsParentheses()
    { return false; };
  virtual ASTType *bindVariable(const char *name, ASTType *value) = 0;
  virtual ASTType *cloneAndResolveIfNamed(TCContext *ctx) 
    { return clone(); };
  virtual ASTType *resolve(TCContext *ctx)
    { return this; };
  virtual void accept(ASTVisitor *worker);
};

/* Just like types, expressions are usually known as ASTExpression* but must sometimes be cast into a more specific
   type, e.g., to invoke sample() on a distribution. In such cases, we call getExpCode() first, which returns
   one of the following values. */

enum expCode { EXP_OTHER, EXP_CONSTNUM, EXP_CONSTSTRING, EXP_TENSOR, EXP_INJ, EXP_AMPERSAND, EXP_FOLD, EXP_DISTRIBUTION, EXP_SET, EXP_CONSTBOOL, EXP_BAG, EXP_LAMBDA, EXP_UNIT };

/* ASTExpression: The base class of all expression nodes. Expressions may be typechecked, cloned, displayed, and 
   evaluated. Evaluation happens by calling 'evalaute' or 'applyTo'. */

class ASTExpression : public ASTBase
{
public:
  ASTExpression(SourceContext *context) : ASTBase(context) {};
  virtual ASTType *getType(TCContext *ctx, TCColons *colons) = 0;
  virtual ASTExpression *clone() { assert(false); return NULL; };
  void showMessage(const char *str);
  virtual ASTExpression *evaluate() = 0;
  virtual expCode getExpCode()
    { return EXP_OTHER; };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) = 0;
  virtual ASTExpression *applyTo(ASTExpression *arg)
    { assert(false); };
  virtual bool isCompileTimeConstant() 
    { return false; };
  virtual void accept(ASTVisitor *worker);
  virtual bool equals(ASTExpression *peer)
    { assert(false); return false; };
  virtual bool needsParentheses()
    { return false; };
};

/* === TYPES ===================================================================================================== */

/* ASTBaseType: One of the built-in types, e.g., 'num' */

class ASTBaseType : public ASTType
{
  char *name;
  
public:
  ASTBaseType(const char *nameArg, SourceContext *context = NULL) : ASTType(context) 
    { this->name = (char*) malloc(strlen(nameArg)+1); strcpy(this->name, nameArg); }
  virtual ~ASTBaseType()
    { free(name); }
  const char *getName()
    { return name; };
  virtual ASTBaseType *clone() 
    { return new ASTBaseType(name, context ? context->clone() : NULL); };
  virtual typeCode getTypeCode()
    { return TYP_BASE; };
  virtual ASTType *bindVariable(const char *name, ASTType *value) 
    { return clone(); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTTypeVariable: A type variable, which may occur in the scope of a mu type. Note the difference to ASTNamedType. */

class ASTTypeVariable : public ASTType
{
  char *name;
  int sequenceNo;
  
public:
  ASTTypeVariable(const char *nameArg, int sequenceNo = 0, SourceContext *context = NULL) : ASTType(context)
    { this->name = (char*) malloc(strlen(nameArg)+1); strcpy(this->name, nameArg); this->sequenceNo = sequenceNo; }
  const char *getName() 
    { return name; };
  int getSequenceNo()
    { return sequenceNo; };
  void setSequenceNo(int newSeq)
    { sequenceNo = newSeq; };
  virtual ASTTypeVariable *clone()
    { return new ASTTypeVariable(name, sequenceNo, context ? context->clone() : NULL); };
  virtual typeCode getTypeCode()
    { return TYP_VAR; };
  virtual ASTType *bindVariable(const char *name, ASTType *value) 
    { return (!strcmp(this->name, name)) ? value->clone() : clone(); };
  virtual void accept(ASTVisitor *worker);
  bool equals(ASTTypeVariable *other)
    { return (!strcmp(name, other->name) && (sequenceNo == other->sequenceNo)); };
};

/* ASTNamedType: A named type created with 'typedef'. Note the difference to ASTTypeVariable. Note that we
   keep the name as long as we can (to avoid confusing the user with the nitty-gritty details of what a 
   list looks like internally) but we have to resolve it, e.g., during equality checking, or at certain
   points during type inference. Don't forget to call 'resolve'! */

class ASTNamedType : public ASTType
{
  char *name;
  
public:
  ASTNamedType(const char *nameArg, SourceContext *context = NULL) : ASTType(context)
    { this->name = (char*) malloc(strlen(nameArg)+1); strcpy(this->name, nameArg); }
  const char *getName() 
    { return name; };
  virtual ASTNamedType *clone()
    { return new ASTNamedType(name, context ? context->clone() : NULL); };
  virtual typeCode getTypeCode()
    { return TYP_NAMED; };
  virtual ASTType *bindVariable(const char *name, ASTType *value) 
    { return clone(); };
  virtual ASTType *cloneAndResolveIfNamed(TCContext *ctx) {
    assert(ctx->containsSymbol(name));
    ASTType *theType = ctx->getSymbolType(name);
    assert(theType);
    return theType->cloneAndResolveIfNamed(ctx);
  }
  virtual ASTType *resolve(TCContext *ctx) { 
    assert(ctx->containsSymbol(name));
    ASTType *theType = ctx->getSymbolType(name);
    assert(theType);
    return theType;
  };
  virtual void accept(ASTVisitor *worker);
};

/* ASTUnionType: The '+' type */

class ASTUnionType : public ASTType
{
  ASTType *option[2];

public:
  ASTUnionType(ASTType *option1, ASTType *option2, SourceContext *context = NULL) : ASTType(context) 
    { this->option[0] = option1; this->option[1] = option2; };
  ASTType *getOption(int i) 
    { assert((i==0) || (i==1)); return option[i]; };
  void setOption(int i, ASTType *typ)
    { assert((i==0) || (i==1)); option[i] = typ; };
  virtual ASTUnionType *clone() 
    { return new ASTUnionType(option[0]->clone(), option[1]->clone(), context ? context->clone() : NULL); };
  virtual typeCode getTypeCode()
    { return TYP_UNION; };
  virtual bool needsParentheses()
    { return true; };
  virtual ASTType *bindVariable(const char *name, ASTType *value) 
    { return new ASTUnionType(option[0]->bindVariable(name, value), option[1]->bindVariable(name, value)); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTSetType: The set type */

class ASTSetType : public ASTType
{
  ASTType *memberType;
  bool isMulti;

public:
  ASTSetType(ASTType *memberType, bool isMulti, SourceContext *context = NULL) : ASTType(context) 
    { this->memberType = memberType; this->isMulti = isMulti; };
  ASTType *getMemberType()
    { return memberType; };
  void setMemberType(ASTType *typ)
    { memberType = typ; };
  virtual ASTSetType *clone() 
    { return new ASTSetType(memberType->clone(), isMulti, context ? context->clone() : NULL); };
  virtual typeCode getTypeCode()
    { return TYP_SET; };
  virtual bool needsParentheses()
    { return true; };
  virtual ASTType *bindVariable(const char *name, ASTType *value) 
    { return new ASTSetType(memberType->bindVariable(name, value), isMulti); };
  virtual void accept(ASTVisitor *worker);
  bool isMultiset() 
    { return isMulti; };
};

/* ASTTensorType: The (X) type */

class ASTTensorType : public ASTType
{
  ASTType *left;
  ASTType *right;

public:
  ASTTensorType(ASTType *left, ASTType *right, SourceContext *context = NULL) : ASTType(context) 
    { this->left = left; this->right = right; };
  ASTType *getLeft() 
    { return left; };
  ASTType *getRight() 
    { return right; };
  void setLeft(ASTType *typ)
    { left = typ; };
  void setRight(ASTType *typ)
    { right = typ; };
  virtual ASTTensorType *clone() 
    { return new ASTTensorType(left->clone(), right->clone(), context ? context->clone() : NULL); };
  virtual typeCode getTypeCode()
    { return TYP_TENSOR; };
  virtual bool needsParentheses()
    { return true; };
  virtual ASTType *bindVariable(const char *name, ASTType *value) 
    { return new ASTTensorType(left->bindVariable(name, value), right->bindVariable(name, value)); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTAmpersandType: The & type */

class ASTAmpersandType : public ASTType
{
  ASTType *left;
  ASTType *right;

public:
  ASTAmpersandType(ASTType *left, ASTType *right, SourceContext *context = NULL) : ASTType(context) 
    { this->left = left; this->right = right; };
  ASTType *getLeft() 
    { return left; };
  ASTType *getRight() 
    { return right; };
  void setLeft(ASTType *typ)
    { left = typ; };
  void setRight(ASTType *typ)
    { right = typ; };
  virtual ASTAmpersandType *clone() 
    { return new ASTAmpersandType(left->clone(), right->clone(), context ? context->clone() : NULL); };
  virtual typeCode getTypeCode()
    { return TYP_AMPERSAND; };
  virtual ASTType *bindVariable(const char *name, ASTType *value) 
    { return new ASTAmpersandType(left->bindVariable(name, value), right->bindVariable(name, value)); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTFuzzyType: The type of a probability distribution */

class ASTFuzzyType : public ASTType
{
  ASTType *innerType;

public:
  ASTFuzzyType(ASTType *innerType, SourceContext *context = NULL) : ASTType(context) 
    { this->innerType = innerType; };
  ASTType *getInnerType() 
    { return innerType; };
  void setInnerType(ASTType *arg) 
    { innerType = arg; };
  virtual ASTFuzzyType *clone() 
    { return new ASTFuzzyType(innerType->clone(), context ? context->clone() : NULL); };
  virtual typeCode getTypeCode()
    { return TYP_FUZZY; };
  virtual ASTType *bindVariable(const char *name, ASTType *value) 
    { return new ASTFuzzyType(innerType->bindVariable(name, value)); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTUnitType: This is basically a 'void' type */

class ASTUnitType : public ASTType
{
public:
  ASTUnitType(SourceContext *context = NULL) : ASTType(context) { };
  virtual typeCode getTypeCode()
    { return TYP_UNIT; };
  virtual ASTUnitType *clone() 
    { return new ASTUnitType(context ? context->clone() : NULL); };
  virtual ASTType *bindVariable(const char *name, ASTType *value) 
    { return clone(); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTFunctionType: Self-explanatory, except for the fact that functions can be annotated with sensitivities
   (different from the paper) */

class ASTFunctionType : public ASTType
{
  ASTType *argType;
  ASTType *resultType;
  double sensitivity;

public:
  ASTFunctionType(ASTType *argType, ASTType *resultType, double sensitivity, SourceContext *context = NULL) : ASTType(context) 
    { this->argType = argType; this->resultType = resultType; this->sensitivity = sensitivity; };
  ASTType *getArgType() 
    { return argType; };
  ASTType *getResultType() 
    { return resultType; };
  void setArgType(ASTType *typ)
    { argType = typ; };
  void setResultType(ASTType *typ)
    { resultType = typ; };
  double getSensitivity() 
    { return sensitivity; };
  virtual ASTFunctionType *clone() 
    { return new ASTFunctionType(argType->clone(), resultType->clone(), sensitivity, context ? context->clone() : NULL); };
  virtual typeCode getTypeCode()
    { return TYP_LAMBDA; };
  virtual ASTType *bindVariable(const char *name, ASTType *value) 
    { return new ASTFunctionType(argType->bindVariable(name, value), resultType->bindVariable(name, value), sensitivity); };
  virtual bool needsParentheses()
    { return true; };
  virtual void accept(ASTVisitor *worker);
};

/* ASTMuType: Recursive types, which are needed, e.g., for lists */

class ASTMuType : public ASTType
{
  char *name;
  ASTType *innerType;
  
public: 
  ASTMuType(const char *name, ASTType *innerType, SourceContext *context = NULL) : ASTType(context)
    { this->name = (char*)malloc(strlen(name)+1); strcpy(this->name, name); this->innerType = innerType; }
  const char *getName() 
    { return name; };
  ASTType *getInnerType() 
    { return innerType; };
  virtual ASTMuType *clone() 
    { return new ASTMuType(name, innerType->clone(), context ? context->clone() : NULL); };
  virtual typeCode getTypeCode()
    { return TYP_MU; };
  virtual ASTType *bindVariable(const char *name, ASTType *value) 
    { return (!strcmp(this->name, name)) ? clone() : new ASTMuType(this->name, innerType->bindVariable(name, value)); };
  ASTType *unroll()
    { return innerType->bindVariable(name, this); };
  virtual bool needsParentheses()
    { return true; };
  virtual void accept(ASTVisitor *worker);
};

/* === EXPRESSIONS =============================================================================================== */

/* ASTVariable: A variable */

class ASTVariable : public ASTExpression
{
  char *name;

public:
  ASTVariable(const char *nameArg, SourceContext *context = NULL) : ASTExpression(context) 
    { this->name = (char*) malloc(strlen(nameArg)+1); strcpy(this->name, nameArg); }
  virtual bool isVariable() 
    { return true; };
  const char *getName() 
    { return name; };
  virtual ASTVariable *clone() 
    { return new ASTVariable(name, context ? context->clone() : NULL); };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *evaluate();
  virtual ASTExpression *substitute(const char *name, ASTExpression *value)
    { if (!strcmp(name, this->name)) return value->clone(); else return clone(); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTCaseExpression: A case expression */

class ASTCaseExpression : public ASTExpression
{
  ASTExpression *expr;
  char *id[2];
  ASTExpression *branch[2];
  
public:
  ASTCaseExpression(ASTExpression *expr, const char *id1, ASTExpression *branch1, const char *id2, ASTExpression *branch2, SourceContext *context = NULL) : ASTExpression(context)
    { this->expr = expr; this->id[0] = (char*)malloc(strlen(id1)); strcpy(this->id[0], id1); this->branch[0] = branch1; this->id[1] = (char*)malloc(strlen(id2)); strcpy(this->id[1], id2); this->branch[1] = branch2; };
  ASTExpression *getExpr() { return expr; };
  const char *getId(int i) { assert((i==0) || (i==1)); return id[i]; };
  ASTExpression *getBranch(int i) { assert((i==0) || (i==1)); return branch[i]; };
  ASTCaseExpression *clone() 
    { return new ASTCaseExpression(expr->clone(), id[0], branch[0]->clone(), id[1], branch[1]->clone(), context ? context->clone() : NULL); };
  virtual ASTExpression *evaluate();
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) { 
    ASTExpression *branch0res = !strcmp(name, id[0]) ? branch[0]->clone() : branch[0]->substitute(name, value);
    ASTExpression *branch1res = !strcmp(name, id[1]) ? branch[1]->clone() : branch[1]->substitute(name, value);
    return new ASTCaseExpression(expr->substitute(name, value), id[0], branch0res, id[1], branch1res, context ? context->clone() : NULL); 
  };
  virtual void accept(ASTVisitor *worker);
}; 
 
/* ASTIfThenElseExpression: A case expression */

class ASTIfThenElseExpression : public ASTExpression
{
  ASTExpression *expr;
  ASTExpression *thenBranch;
  ASTExpression *elseBranch;
  
public:
  ASTIfThenElseExpression(ASTExpression *expr, ASTExpression *thenBranch, ASTExpression *elseBranch, SourceContext *context = NULL) : ASTExpression(context)
    { this->expr = expr; this->thenBranch = thenBranch; this->elseBranch = elseBranch; };
  ASTExpression *getExpr() { return expr; };
  ASTExpression *getThenBranch() { return thenBranch; };
  ASTExpression *getElseBranch() { return elseBranch; };
  ASTIfThenElseExpression *clone() 
    { return new ASTIfThenElseExpression(expr->clone(), thenBranch->clone(), elseBranch->clone(), context ? context->clone() : NULL); };
  virtual ASTExpression *evaluate();
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) { 
    return new ASTIfThenElseExpression(expr->substitute(name, value), thenBranch->substitute(name, value), elseBranch->substitute(name, value), context ? context->clone() : NULL); 
  };
  virtual void accept(ASTVisitor *worker);
}; 
 
/* ASTInfixOpExpression: Used for functions like addition or multiplication that take two arguments */
 
class ASTInfixOpExpression : public ASTExpression
{
protected:
  ASTExpression *left;
  ASTExpression *right;
  const char *op;

public:
  ASTInfixOpExpression(ASTExpression *left, const char *op, ASTExpression *right, SourceContext *context = NULL) : ASTExpression(context)
    { this->left = left; this->op = op; this->right = right; }
  ASTExpression *getLeftExpression() { return left; };
  ASTExpression *getRightExpression() { return right; };
  const char *getOp() { return op; };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *evaluate();
  ASTInfixOpExpression *clone() 
    { return new ASTInfixOpExpression(left->clone(), op, right->clone(), context ? context->clone() : NULL); };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return new ASTInfixOpExpression(left->substitute(name, value), op, right->substitute(name, value), context ? context->clone() : NULL); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTUnaryExpression: A base class for many unary operators */

class ASTUnaryExpression : public ASTExpression
{
protected:
  ASTExpression *innerExp;

public:
  ASTUnaryExpression(ASTExpression *innerExp, SourceContext *context) : ASTExpression(context)
    { this->innerExp = innerExp; }
  ASTExpression *getInnerExp() { return innerExp; };
  virtual void accept(ASTVisitor *worker);
};

/* ASTUnfoldExpression: The unfold operator */

class ASTUnfoldExpression : public ASTUnaryExpression
{
public:
  ASTUnfoldExpression(ASTExpression *innerExp, SourceContext *context = NULL) : ASTUnaryExpression(innerExp, context) {};
  virtual ASTExpression *evaluate();
  virtual ASTExpression *clone() 
    { return new ASTUnfoldExpression(innerExp->clone(), context ? context->clone() : NULL); };
  ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return new ASTUnfoldExpression(innerExp->substitute(name, value), context ? context->clone() : NULL); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTBuiltinExpression: Used for built-in functions like 'return' or 'fuzz' to avoid having to define
   a special class for each of them. The 'applyTo' function of this class is basically a big
   switch statement, which applies the appropriate function depending on the 'name' field. */

class ASTBuiltinExpression : public ASTExpression
{
  static const int maxNumArgs = 4;
  ASTExpression *myArg[maxNumArgs];
  int numArgs;
  const char *name;

public:
  ASTBuiltinExpression(const char *name, int numArgs = 0, ASTExpression **args = NULL, SourceContext *context = NULL) : ASTExpression(context) 
    { this->name = name; this->numArgs = numArgs;for (int i=0; i<maxNumArgs; i++) myArg[i] = (i<numArgs) ? args[i] : NULL; };
  void addArgument(ASTExpression *newArg)
    { assert(numArgs < maxNumArgs); myArg[numArgs++] = newArg; };
  virtual ASTExpression *evaluate();
  virtual ASTBuiltinExpression *clone()
    { ASTBuiltinExpression *retval = new ASTBuiltinExpression(name, 0, NULL, context ? context->clone() : NULL); for (int i=0; i<numArgs; i++) retval->addArgument(myArg[i]->clone()); return retval; };
  ASTType *getType(TCContext *ctx, TCColons *colons)
    { assert(false); return NULL; };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return clone(); };
  virtual ASTExpression *applyTo(ASTExpression *arg);
  virtual void accept(ASTVisitor *worker);
  const char *getName()
    { return name; };
};

/* ASTUnaryExpressionWithType: A base class for unary operators that have a type annotation */

class ASTUnaryExpressionWithType : public ASTUnaryExpression
{
protected:
  ASTType *typeArg;

public:
  ASTUnaryExpressionWithType(ASTExpression *innerExp, ASTType *typeArg, SourceContext *context) : ASTUnaryExpression(innerExp, context)
    { this->innerExp = innerExp; this->typeArg = typeArg; }
  ASTType *getTypeArg() { return typeArg; };
  virtual void accept(ASTVisitor *worker);
};

/* ASTFoldExpression: Self-explanatory */

class ASTFoldExpression : public ASTUnaryExpressionWithType
{
public:
  ASTFoldExpression(ASTExpression *innerExp, ASTType *typeArg, SourceContext *context = NULL) : ASTUnaryExpressionWithType(innerExp, typeArg, context) {};
  virtual ASTExpression *evaluate();
  virtual ASTExpression *clone() 
    { return new ASTFoldExpression(innerExp->clone(), typeArg->clone(), context ? context->clone() : NULL); };
  ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual expCode getExpCode()
    { return EXP_FOLD; };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return new ASTFoldExpression(innerExp->substitute(name, value), typeArg->clone(), context ? context->clone() : NULL); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTInjectExpression: Self-explanatory */

class ASTInjectExpression : public ASTUnaryExpressionWithType
{
  int branch;
  
public:
  ASTInjectExpression(int branch0or1, ASTExpression *innerExp, ASTType *typeArg, SourceContext *context = NULL) : ASTUnaryExpressionWithType(innerExp, typeArg, context)
    { this->branch = branch0or1; }
  int getBranch() { return branch; };
  virtual ASTExpression *evaluate();
  virtual ASTExpression *clone() 
    { return new ASTInjectExpression(branch, innerExp->clone(), typeArg->clone(), context ? context->clone() : NULL); };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual expCode getExpCode()
    { return EXP_INJ; };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return new ASTInjectExpression(branch, innerExp->substitute(name, value), typeArg->clone(), context ? context->clone() : NULL); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTProjectExpression: This is for 'fst' and 'snd'. */

class ASTProjectExpression : public ASTUnaryExpression
{
  int branch;
  
public:
  ASTProjectExpression(int branch0or1, ASTExpression *innerExp, SourceContext *context = NULL) : ASTUnaryExpression(innerExp, context)
    { this->branch = branch0or1; }
  int getBranch() 
    { return branch; };
  virtual ASTExpression *evaluate();
  virtual ASTProjectExpression *clone() 
    { return new ASTProjectExpression(branch, innerExp->clone(), context ? context->clone() : NULL); };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return new ASTProjectExpression(branch, innerExp->substitute(name, value), context ? context->clone() : NULL); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTTensorTuple: A tensor tuple. Should really be called TensorTupleExpression. */

class ASTTensorTuple : public ASTExpression
{
protected:
  ASTExpression *left;
  ASTExpression *right;

public:
  ASTTensorTuple(ASTExpression *left, ASTExpression *right, SourceContext *context = NULL) : ASTExpression(context)
    { this->left = left; this->right = right; }
  ASTExpression *getLeftExpression() { return left; };
  ASTExpression *getRightExpression() { return right; };
  void setRightExpression(ASTExpression *exp) { this->right = exp; };
  virtual ASTExpression *evaluate();
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *clone() 
    { return new ASTTensorTuple(left->clone(), right->clone(), context ? context->clone() : NULL); };
  virtual expCode getExpCode()
    { return EXP_TENSOR; };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return new ASTTensorTuple(left->substitute(name, value), right->substitute(name, value), context ? context->clone() : NULL); };
  virtual void accept(ASTVisitor *worker);
  virtual bool equals(ASTExpression *peer)
    { return (peer->getExpCode() == EXP_TENSOR) && ((ASTTensorTuple*)peer)->getLeftExpression()->equals(left) && ((ASTTensorTuple*)peer)->getRightExpression()->equals(right); };
};

/* ASTTensorDestructorExpression: A tensor destructor */

class ASTTensorDestructorExpression : public ASTExpression
{
protected:
  char *name1;
  char *name2;
  ASTExpression *producingExp;
  ASTExpression *consumingExp;

public:
  ASTTensorDestructorExpression(const char *name1, const char *name2, ASTExpression *producingExp, ASTExpression *consumingExp, SourceContext *context = NULL) : ASTExpression(context)
    { this->name1 = (char*)malloc(strlen(name1)); strcpy(this->name1, name1); this->name2 = (char*)malloc(strlen(name2)); strcpy(this->name2, name2); this->producingExp = producingExp; this->consumingExp = consumingExp; }
  const char *getName(int i) { assert((i==0) || (i==1)); return i ? name2 : name1; };
  ASTExpression *getProducingExp() { return producingExp; };
  ASTExpression *getConsumingExp() { return consumingExp; };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *evaluate();
  virtual ASTExpression *clone() 
    { return new ASTTensorDestructorExpression(name1, name2, producingExp->clone(), consumingExp->clone(), context ? context->clone() : NULL); };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) {
    ASTExpression *consumingSubst = (!strcmp(name, name1) || !strcmp(name, name2)) ? consumingExp->clone() : consumingExp->substitute(name, value);
    return new ASTTensorDestructorExpression(name1, name2, producingExp->substitute(name, value), consumingSubst, context ? context->clone() : NULL);
  };
  virtual void accept(ASTVisitor *worker);
};

/* ASTAmpersandTuple: An ampersand tuple */

class ASTAmpersandTuple : public ASTExpression
{
protected:
  ASTExpression *left;
  ASTExpression *right;

public:
  ASTAmpersandTuple(ASTExpression *left, ASTExpression *right, SourceContext *context = NULL) : ASTExpression(context)
    { this->left = left; this->right = right; }
  ASTExpression *getLeftExpression() { return left; };
  ASTExpression *getRightExpression() { return right; };
  virtual ASTExpression *evaluate();
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *clone() 
    { return new ASTAmpersandTuple(left->clone(), right->clone(), context ? context->clone() : NULL); };
  virtual expCode getExpCode()
    { return EXP_AMPERSAND; };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return new ASTAmpersandTuple(left->substitute(name, value), right->substitute(name, value), context ? context->clone() : NULL); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTLetBindingExpression: A let binding with an optional type assertion */

class ASTLetBindingExpression : public ASTExpression
{
protected:
  char *nameToBind;
  ASTExpression *rightHandExp;
  ASTExpression *followExp;
  ASTType *assertedTypeOrNull;
  
public:
  ASTLetBindingExpression(const char *nameToBind, ASTExpression *rightHandExp, ASTExpression *followExp, ASTType *assertedTypeOrNull, SourceContext *context = NULL) : ASTExpression(context)
    { this->rightHandExp = rightHandExp; this->nameToBind = (char*) malloc(strlen(nameToBind)+1); strcpy(this->nameToBind, nameToBind); this->assertedTypeOrNull = assertedTypeOrNull; this->followExp = followExp; }
  const char *getNameToBind() { return nameToBind; }
  ASTExpression *getRightHandExp() { return rightHandExp; };
  ASTExpression *getFollowExp() { return followExp; };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *evaluate();
  virtual ASTType *getAssertedTypeOrNull()
    { return assertedTypeOrNull; };
  virtual ASTLetBindingExpression *clone() 
    { return new ASTLetBindingExpression(nameToBind, rightHandExp->clone(), followExp->clone(), assertedTypeOrNull ? assertedTypeOrNull->clone() : NULL, context ? context->clone() : NULL); };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) {
    if (!strcmp(nameToBind, name))
      return new ASTLetBindingExpression(nameToBind, rightHandExp->substitute(name, value), followExp->clone(), assertedTypeOrNull ? assertedTypeOrNull->clone() : NULL, context ? context->clone() : NULL);
    return new ASTLetBindingExpression(nameToBind, rightHandExp->substitute(name, value), followExp->substitute(name, value), assertedTypeOrNull ? assertedTypeOrNull->clone() : NULL, context ? context->clone() : NULL);
  };
  virtual void accept(ASTVisitor *worker);
};

/* ASTPrintExpression: Debugging node */

class ASTPrintExpression : public ASTExpression
{
protected:
  ASTExpression *rightHandExp;
  ASTExpression *followExp;
  
public:
  ASTPrintExpression(ASTExpression *rightHandExp, ASTExpression *followExp, SourceContext *context = NULL) : ASTExpression(context)
    { this->rightHandExp = rightHandExp; this->followExp = followExp; }
  ASTExpression *getRightHandExp() { return rightHandExp; };
  ASTExpression *getFollowExp() { return followExp; };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *evaluate();
  virtual ASTPrintExpression *clone() 
    { return new ASTPrintExpression(rightHandExp->clone(), followExp->clone(), context ? context->clone() : NULL); };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) {
    return new ASTPrintExpression(rightHandExp->substitute(name, value), followExp->substitute(name, value), context ? context->clone() : NULL);
  };
  virtual void accept(ASTVisitor *worker);
};

/* ASTSampleExpression: A special kind of let-binding that samples from a distribution */

class ASTSampleExpression : public ASTExpression
{
protected:
  char *nameToBind;
  ASTExpression *rightHandExp;
  ASTExpression *followExp;
  
public:
  ASTSampleExpression(const char *nameToBind, ASTExpression *rightHandExp, ASTExpression *followExp, SourceContext *context = NULL) : ASTExpression(context)
    { this->rightHandExp = rightHandExp; this->nameToBind = (char*) malloc(strlen(nameToBind)+1); strcpy(this->nameToBind, nameToBind); this->followExp = followExp; }
  const char *getNameToBind() { return nameToBind; }
  ASTExpression *getRightHandExp() { return rightHandExp; };
  ASTExpression *getFollowExp() { return followExp; };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *evaluate();
  virtual ASTSampleExpression *clone() 
    { return new ASTSampleExpression(nameToBind, rightHandExp->clone(), followExp->clone(), context ? context->clone() : NULL); };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) {
    if (!strcmp(nameToBind, name))
      return new ASTSampleExpression(nameToBind, rightHandExp->substitute(name, value), followExp->clone(), context ? context->clone() : NULL);
    return new ASTSampleExpression(nameToBind, rightHandExp->substitute(name, value), followExp->substitute(name, value), context ? context->clone() : NULL);
  };
  virtual void accept(ASTVisitor *worker);
};

/* ASTTypeDefExpression: A typedef. This does nothing when applied, but it does define a type during type checking */

class ASTTypeDefExpression : public ASTExpression
{
protected:
  char *nameToBind;
  ASTType *typeDef;
  ASTExpression *followExp;
  
public:
  ASTTypeDefExpression(const char *nameToBind, ASTType *typeDef, ASTExpression *followExp, SourceContext *context = NULL) : ASTExpression(context)
    { this->typeDef = typeDef; this->nameToBind = (char*) malloc(strlen(nameToBind)+1); strcpy(this->nameToBind, nameToBind); this->followExp = followExp; }
  const char *getNameToBind() { return nameToBind; }
  ASTType *getTypeDef() { return typeDef; };
  ASTExpression *getFollowExp() { return followExp; };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *evaluate();
  virtual ASTTypeDefExpression *clone() 
    { return new ASTTypeDefExpression(nameToBind, typeDef->clone(), followExp->clone(), context ? context->clone() : NULL); };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return new ASTTypeDefExpression(nameToBind, typeDef->clone(), followExp->substitute(name, value), context ? context->clone() : NULL); };
  virtual void accept(ASTVisitor *worker);
};

/* ASTLambdaExpression: A possibly recursive lambda expression. If recursive, it needs to know its own name,
   as well as its return type. */

class ASTLambdaExpression : public ASTExpression
{
protected:
  ASTExpression *innerExp;
  ASTType *bindingType;
  double declaredSensitivity;
  char *nameToBind;
  char *ownNameOrNull;
  ASTType *returnTypeOrNull;

public:
  ASTLambdaExpression(const char *nameToBind, ASTExpression *innerExp, ASTType *bindingType, double declaredSensitivity, const char *ownNameOrNull, ASTType *returnTypeOrNull, SourceContext *context = NULL) : ASTExpression(context)
    { this->innerExp = innerExp; this->nameToBind = (char*) malloc(strlen(nameToBind)+1); this->bindingType = bindingType; strcpy(this->nameToBind, nameToBind); this->declaredSensitivity = declaredSensitivity; if (ownNameOrNull) { this->ownNameOrNull = (char*) malloc(strlen(ownNameOrNull)+1); strcpy(this->ownNameOrNull, ownNameOrNull); } else { this->ownNameOrNull = NULL; } this->returnTypeOrNull = returnTypeOrNull; }
  const char *getNameToBind() { return nameToBind; }
  ASTType *getBindingType() { return bindingType; };
  double getDeclaredSensitivity() { return declaredSensitivity; };
  ASTType *getReturnTypeOrNull() { return returnTypeOrNull; };
  const char *getOwnNameOrNull() { return ownNameOrNull; };
  ASTExpression *getInnerExp() { return innerExp; };
  virtual expCode getExpCode() { return EXP_LAMBDA; };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *evaluate();
  virtual ASTLambdaExpression *clone() 
    { return new ASTLambdaExpression(nameToBind, innerExp->clone(), bindingType->clone(), declaredSensitivity, ownNameOrNull, returnTypeOrNull ? returnTypeOrNull->clone() : NULL, context ? context->clone() : NULL); };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) {
    ASTExpression *innerSubst = (!strcmp(nameToBind, name) || (ownNameOrNull && !strcmp(ownNameOrNull, name))) ? innerExp->clone() : innerExp->substitute(name, value);
    return new ASTLambdaExpression(nameToBind, innerSubst, bindingType->clone(), declaredSensitivity, ownNameOrNull, returnTypeOrNull ? returnTypeOrNull->clone() : NULL, context ? context->clone() : NULL);
  };
  virtual ASTExpression *applyTo(ASTExpression *arg);
  virtual void accept(ASTVisitor *worker);
};

/* ASTApplyExpression: Applies a function to another expression */

class ASTApplyExpression : public ASTExpression
{
protected:
  ASTExpression *fun;
  ASTExpression *arg;

public:
  ASTApplyExpression(ASTExpression *fun, ASTExpression *arg, SourceContext *context = NULL) : ASTExpression(context)
    { this->fun = fun; this->arg = arg; }
  ASTExpression *getFun() { return fun; };
  ASTExpression *getArg() { return arg; };
  virtual ASTExpression *evaluate();
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *clone() 
    { return new ASTApplyExpression(fun->clone(), arg->clone(), context ? context->clone() : NULL); };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return new ASTApplyExpression(fun->substitute(name, value), arg->substitute(name, value), context ? context->clone() : NULL); };
  virtual void accept(ASTVisitor *worker);
  virtual bool needsParentheses()
    { return true; };
};

/* ASTUnitValue: The one value of the unit type. */

class ASTUnitValue : public ASTExpression
{
public:
  ASTUnitValue(SourceContext *context = NULL) : ASTExpression(context) {};
  virtual ASTExpression *evaluate();
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *clone() 
    { return new ASTUnitValue(context ? context->clone() : NULL); };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return clone(); };
  virtual expCode getExpCode()
    { return EXP_UNIT; };
  virtual void accept(ASTVisitor *worker);
};

/* ASTConstExpression: The base class for constant expressions */

class ASTConstExpression : public ASTExpression
{
public:
  ASTConstExpression(SourceContext *context) : ASTExpression(context) {};
  virtual bool isCompileTimeConstant() { return true; };
  virtual double evaluateAtCompileTime() = 0;
  virtual void accept(ASTVisitor *worker);
};

/* ASTDistribution: The base class for distributions (Laplace, single-value, ...) that can be sampled. */

class ASTDistribution : public ASTExpression {
public:
  ASTDistribution(SourceContext *context = NULL) : ASTExpression(context) 
    { };
  virtual expCode getExpCode()
    { return EXP_DISTRIBUTION; };
  virtual ASTExpression *sample() = 0;
};

class ASTSetExpression : public ASTExpression {
protected:
  ASTType *memberType;
  bool isMulti;
  
public:
  struct elementStruct {
    ASTExpression *value;
    struct elementStruct *next;
  } *elements;

  ASTSetExpression(ASTType *memberType, bool isMulti, SourceContext *context = NULL);
  ASTType *getMemberType()
    { return memberType; };
  virtual expCode getExpCode()
    { return isMulti ? EXP_BAG : EXP_SET; };
  void add(ASTExpression *val);
  bool contains(ASTExpression *val);
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *evaluate()
    { return clone(); };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return clone(); };
  virtual void accept(ASTVisitor *worker);
  virtual ASTExpression *clone();
  int count();
};

/* ASTConstNumber: A numeric constant. */

class ASTConstNumber : public ASTConstExpression
{
protected:
  double value;

public:
  ASTConstNumber(double value, SourceContext *context = NULL) : ASTConstExpression(context) { this->value = value; };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *evaluate();
  double getValue()
    { return value; };
  virtual ASTExpression *clone() 
    { return new ASTConstNumber(value, context ? context->clone() : NULL); };
  virtual expCode getExpCode()
    { return EXP_CONSTNUM; };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return clone(); };
  virtual void accept(ASTVisitor *worker);
  virtual double evaluateAtCompileTime()
    { return value; };
  virtual bool equals(ASTExpression *peer)
    { return (peer->getExpCode() == EXP_CONSTNUM) && (((ASTConstNumber*)peer)->getValue() == value); };
};

/* ASTConstString: A string constant. */

class ASTConstString : public ASTExpression
{
protected:
  char *value;

public:
  ASTConstString(const char *value, SourceContext *context = NULL) : ASTExpression(context) 
    { this->value = (char*)malloc(strlen(value)+1); strcpy(this->value, value); };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *evaluate();
  const char *getValue()
    { return value; };
  virtual ASTExpression *clone() 
    { return new ASTConstString(value, context ? context->clone() : NULL); };
  virtual expCode getExpCode()
    { return EXP_CONSTSTRING; };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return clone(); };
  virtual void accept(ASTVisitor *worker);
  virtual bool equals(ASTExpression *peer)
    { return (peer->getExpCode() == EXP_CONSTSTRING) && !strcmp(((ASTConstString*)peer)->getValue(), this->value); };
};

/* ASTConstBool: A boolean constant. */

class ASTConstBool : public ASTExpression
{
protected:
  bool value;

public:
  ASTConstBool(bool value, SourceContext *context = NULL) : ASTExpression(context) 
    { this->value = value; };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons);
  virtual ASTExpression *evaluate();
  bool getValue()
    { return value; };
  virtual ASTExpression *clone() 
    { return new ASTConstBool(value, context ? context->clone() : NULL); };
  virtual expCode getExpCode()
    { return EXP_CONSTBOOL; };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return clone(); };
  virtual void accept(ASTVisitor *worker);
  virtual bool equals(ASTExpression *peer)
    { return (peer->getExpCode() == EXP_CONSTBOOL) && (((ASTConstBool*)peer)->getValue() == value); };
};

/* ASTSingleValueDistribution: A distribution that always returns the same value when sampled.
   Needed to implement the 'return' operator. */

class ASTSingleValueDistribution : public ASTDistribution {
protected:
  ASTExpression *exp;

public:
  ASTSingleValueDistribution(ASTExpression *exp, SourceContext *context = NULL) : ASTDistribution(context)
    { this->exp = exp; };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons) 
    { assert(false); };
  virtual ASTExpression *evaluate()
    { return clone(); };
  virtual ASTSingleValueDistribution *clone() 
    { return new ASTSingleValueDistribution(exp->clone(), context ? context->clone() : NULL); };
  virtual ASTExpression *sample()
    { return exp->clone(); };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return clone(); };
  ASTExpression *getCenter()
    { return exp; };
  virtual void accept(ASTVisitor *worker);
};

extern double epsilon;
extern bool disableFuzz;
double laplaceDraw(double mu, double b);

/* ASTLaplaceDistribution: A Laplace distribution, centered around a specific value. */

class ASTLaplaceDistribution : public ASTDistribution {
protected:
  double val;

public:
  ASTLaplaceDistribution(double val, SourceContext *context = NULL) : ASTDistribution(context)
    { this->val = val; };
  virtual ASTType *getType(TCContext *ctx, TCColons *colons) 
    { assert(false); };
  virtual ASTExpression *evaluate()
    { return clone(); };
  virtual ASTLaplaceDistribution *clone() 
    { return new ASTLaplaceDistribution(val, context ? context->clone() : NULL); };
  virtual ASTExpression *sample()
    { return new ASTConstNumber(disableFuzz ? val : laplaceDraw(val, 1/epsilon)); };
  virtual ASTExpression *substitute(const char *name, ASTExpression *value) 
    { return clone(); };
  double getCenter()
    { return val; };
  virtual void accept(ASTVisitor *worker);
};  

/* === VISITOR =================================================================================================== */

class ASTVisitor 
{
public:
  virtual void visit(ASTBase *peer) { assert(false); };
  virtual void visit(ASTType *peer) { assert(false); };
  virtual void visit(ASTBaseType *peer) { assert(false); };
  virtual void visit(ASTTypeVariable *peer) { assert(false); };
  virtual void visit(ASTUnionType *peer) { assert(false); };
  virtual void visit(ASTTensorType *peer) { assert(false); };
  virtual void visit(ASTAmpersandType *peer) { assert(false); };
  virtual void visit(ASTFuzzyType *peer) { assert(false); };
  virtual void visit(ASTUnitType *peer) { assert(false); };
  virtual void visit(ASTFunctionType *peer) { assert(false); };
  virtual void visit(ASTMuType *peer) { assert(false); };
  virtual void visit(ASTExpression *peer) { assert(false); };
  virtual void visit(ASTVariable *peer) { assert(false); };
  virtual void visit(ASTCaseExpression *peer) { assert(false); };
  virtual void visit(ASTInfixOpExpression *peer) { assert(false); };
  virtual void visit(ASTUnaryExpression *peer) { assert(false); };
  virtual void visit(ASTUnfoldExpression *peer) { assert(false); };
  virtual void visit(ASTBuiltinExpression *peer) { assert(false); };
  virtual void visit(ASTUnaryExpressionWithType *peer) { assert(false); };
  virtual void visit(ASTFoldExpression *peer) { assert(false); };
  virtual void visit(ASTInjectExpression *peer) { assert(false); };
  virtual void visit(ASTProjectExpression *peer) { assert(false); };
  virtual void visit(ASTTensorTuple *peer) { assert(false); };
  virtual void visit(ASTTensorDestructorExpression *peer) { assert(false); };
  virtual void visit(ASTAmpersandTuple *peer) { assert(false); };
  virtual void visit(ASTLetBindingExpression *peer) { assert(false); };
  virtual void visit(ASTLambdaExpression *peer) { assert(false); };
  virtual void visit(ASTApplyExpression *peer) { assert(false); };
  virtual void visit(ASTUnitValue *peer) { assert(false); };
  virtual void visit(ASTConstExpression *peer) { assert(false); };
  virtual void visit(ASTConstNumber *peer) { assert(false); };
  virtual void visit(ASTTypeDefExpression *peer) { assert(false); };
  virtual void visit(ASTNamedType *peer) { assert(false); };
  virtual void visit(ASTSetType *peer) { assert(false); };
  virtual void visit(ASTSampleExpression *peer) { assert(false); };
  virtual void visit(ASTSetExpression *peer) { assert(false); };
  virtual void visit(ASTSingleValueDistribution *peer) { assert(false); };
  virtual void visit(ASTLaplaceDistribution *peer) { assert(false); };
  virtual void visit(ASTConstBool *peer) { assert(false); };
  virtual void visit(ASTIfThenElseExpression *peer) { assert(false); };
  virtual void visit(ASTPrintExpression *peer) { assert(false); };
  virtual void visit(ASTConstString *peer) { assert(false); };
};

class ASTDFSVisitor : public ASTVisitor
{
public:
  ASTDFSVisitor() : ASTVisitor() {}
  virtual void visit(ASTBaseType *peer) { };
  virtual void visit(ASTTypeVariable *peer) { };
  virtual void visit(ASTUnionType *peer) { peer->getOption(0)->accept(this); peer->getOption(1)->accept(this); };
  virtual void visit(ASTTensorType *peer) { peer->getLeft()->accept(this); peer->getRight()->accept(this); };
  virtual void visit(ASTAmpersandType *peer) { peer->getLeft()->accept(this); peer->getRight()->accept(this); };
  virtual void visit(ASTFuzzyType *peer) { peer->getInnerType()->accept(this); };
  virtual void visit(ASTUnitType *peer) { };
  virtual void visit(ASTFunctionType *peer) { peer->getArgType()->accept(this); peer->getResultType()->accept(this); };
  virtual void visit(ASTMuType *peer) { peer->getInnerType()->accept(this); };
  virtual void visit(ASTVariable *peer) { };
  virtual void visit(ASTCaseExpression *peer) { peer->getExpr()->accept(this); peer->getBranch(0)->accept(this); peer->getBranch(1)->accept(this); };
  virtual void visit(ASTInfixOpExpression *peer) { peer->getLeftExpression()->accept(this); peer->getRightExpression()->accept(this); };
  virtual void visit(ASTUnfoldExpression *peer) { peer->getInnerExp()->accept(this); };
  virtual void visit(ASTBuiltinExpression *peer) { };
  virtual void visit(ASTFoldExpression *peer) { peer->getInnerExp()->accept(this); peer->getTypeArg()->accept(this); };
  virtual void visit(ASTInjectExpression *peer) { peer->getInnerExp()->accept(this); peer->getTypeArg()->accept(this); };
  virtual void visit(ASTProjectExpression *peer) { peer->getInnerExp()->accept(this); };
  virtual void visit(ASTTensorTuple *peer) { peer->getLeftExpression()->accept(this); peer->getRightExpression()->accept(this); };
  virtual void visit(ASTTensorDestructorExpression *peer) { peer->getProducingExp()->accept(this); peer->getConsumingExp()->accept(this); };
  virtual void visit(ASTAmpersandTuple *peer) { peer->getLeftExpression()->accept(this); peer->getRightExpression()->accept(this); };
  virtual void visit(ASTLetBindingExpression *peer) { peer->getRightHandExp()->accept(this); peer->getFollowExp()->accept(this); if (peer->getAssertedTypeOrNull()) peer->getAssertedTypeOrNull()->accept(this); };
  virtual void visit(ASTLambdaExpression *peer) { peer->getInnerExp()->accept(this); peer->getBindingType()->accept(this); if (peer->getReturnTypeOrNull()) peer->getReturnTypeOrNull()->accept(this); };
  virtual void visit(ASTApplyExpression *peer) { peer->getFun()->accept(this); peer->getArg()->accept(this); };
  virtual void visit(ASTUnitValue *peer) { };
  virtual void visit(ASTConstNumber *peer) { };
  virtual void visit(ASTTypeDefExpression *peer) { peer->getTypeDef()->accept(this); peer->getFollowExp()->accept(this); };
  virtual void visit(ASTNamedType *peer) { };
  virtual void visit(ASTSetType *peer) { peer->getMemberType()->accept(this); };
  virtual void visit(ASTSampleExpression *peer) { peer->getRightHandExp()->accept(this); peer->getFollowExp()->accept(this); };
  virtual void visit(ASTSetExpression *peer) { assert(false); };
  virtual void visit(ASTIfThenElseExpression *peer) { peer->getExpr()->accept(this); peer->getThenBranch()->accept(this); peer->getElseBranch()->accept(this); };
  virtual void visit(ASTPrintExpression *peer) { peer->getRightHandExp()->accept(this); peer->getFollowExp()->accept(this); };
  virtual void visit(ASTConstString *peer) { };
};

#endif /* defined(__il_h__) */
