#ifndef __frontend_h__
#define __frontend_h__

#define BUFSIZE 500

#include <unistd.h>

#include "ast.h"

enum SymbolClass { SYM_LETBINDING, SYM_TYPEVAR, SYM_TYPENAME };

class FEScope
{
protected:
  struct symbolRecord {
    struct symbolRecord *next;
    SymbolClass cls;
    char *name;
  } *symbols;

  FEScope *parentScope;

  bool symbolExistsLocally(const char *identifier, SymbolClass cls);
  
public:
  FEScope(FEScope *parentScope, SourceContext *context)
    { this->parentScope = parentScope; this->symbols = NULL; };
    
  bool symbolExists(const char *identifier, SymbolClass cls);
  bool symbolExists(const char *identifier)
    { return symbolExists(identifier, SYM_LETBINDING) || symbolExists(identifier, SYM_TYPEVAR) || symbolExists(identifier, SYM_TYPENAME); };
  void add(SymbolClass cls, const char *id);
  FEScope *getParentScope() { return parentScope; };
};

class FEParameter 
{
protected:
  char *name;
  double sensitivity;
  ASTType *type;
  FEParameter *next;
  
public:
  FEParameter(const char *nameArg, double sensitivity, ASTType *type, SourceContext *context)
    { this->next = NULL; this->sensitivity = sensitivity; this->type = type; this->name = (char*) malloc(strlen(nameArg)+1); strcpy(this->name, nameArg); }
  void append(FEParameter *next)
    { if (this->next) this->next->append(next); else this->next = next; };
  const char *getName()
    { return name; };
  double getSensitivity()
    { return sensitivity; };
  ASTType *getType()
    { return type; };
  FEParameter *getNext()
    { return next; };
};

class InputContext
{
protected:
  FEScope *currentScope;
  const char *cppPath;
  FILE *input;
  bool fail;
  bool isPrimaryError;
  int errorcount;
  int warningcount;
  int pipe_end[2];
  pid_t cppPid;
  const char *options;

  void ShowErrorPos();

public:
  const char *filename;
  char linebuf[BUFSIZE];
  int lineno;
  int tokenPos;

  InputContext(const char *filename);

  void Error(const char *fmt, ...);
  void Warning(const char *fmt, ...);
  bool hasErrors() { return fail; };
  FILE *getFile() { return input; }
  int getErrorCount() { return errorcount; }
  int getWarningCount() { return warningcount; }
  void NextLine() { lineno++; }
  void rememberError(const char *linebuf, int tokenPos);
  const char *getFilename() { return filename; }
  void setPosition(const char *filename, int lineno) { this->filename = filename; this->lineno = lineno; }
  void setOptions(const char *options) { this->options = options; };
  void update(InputContext *c) { errorcount += c->errorcount; warningcount += c->warningcount; if (c->fail) fail=true; }
  void openPipe();
  void closePipe();
  void setCppPath(const char *cppPath) { this->cppPath = cppPath; }
  ASTExpression *parse(FEScope *rootScope);
  FEScope *getCurrentScope() { return currentScope; }
  void setCurrentScope(FEScope *scope) { currentScope = scope; }
};

extern InputContext *currentContext;

#endif /* defined(__frontend_h__) */
