#ifndef __typechecker_h__
#define __typechecker_h__

#include <string.h>

class ASTType;
class ASTExpression;
class ASTTypeVariable;
class ASTVisitor;
class TCContext;

/* TCVariableMapping: Contains a mapping from type variables in one context to type variables in another. 
   This is necessary when testing types for equality. */ 

class TCVariableMapping {
protected:
  struct mapEntry { // not circular
    char *name;
    char *mappedName;
    struct mapEntry *next;
    struct mapEntry *prev;
  } *maps;
  
public:
  TCVariableMapping()
    { this->maps = NULL; };
  ~TCVariableMapping()
    { while (maps) removeMapping(maps->name); };
  void addMapping(const char *name, const char *mappedName);
  void removeMapping(const char *name);
  bool mappingExists(const char *name) { return lookupMapping(name) != NULL; };
  const char *lookupMapping(const char *name);
};

/* TCSymbol: Represents a symbol in the context (TCContext) used for type checking. */

class TCSymbol {
protected:
  TCSymbol *prev;
  TCSymbol *next;
  char *name;
  ASTType *type;
  int handle;
  
public:
  TCSymbol(const char *name, ASTType *type, int handle = 0);
  ~TCSymbol();
  const char *getName() 
    { return name; };
  ASTType *getType() 
    { return type; };
  void setType(ASTType *typ)
    { this->type = typ; };
  TCSymbol *lookup(const char *name, TCSymbol *last);
  TCSymbol *lookupByHandle(int handle, TCSymbol *last);
  void enqueueBefore(TCSymbol *sym);
  char *render(char *buf, TCSymbol *last, int maxlen, bool first = true);
  TCSymbol *getNext() 
    { return next; };
  void dequeue();
};  

class TCColons {
protected:
  struct colonEntry {
    char *name;
    double colon;
    struct colonEntry *next;
  } *entries;
  
  struct colonEntry *findSymbol(const char *name);
  
public:
  TCColons();
  void addSymbol(const char *name, double sensitivity = 0);
  void setSymbol(const char *name, double sensitivity);
  double getSymbol(const char *name);
  bool containsSymbol(const char *name);
  void removeSymbol(const char *name);
  void add(TCColons *peer);
  void multiplyBy(double factor);
  void takeMax(TCColons *peer);
  bool containsNonzero();
  char *render();
};

/* TCContext: The context used for type checking. */

class TCContext {
protected:
  TCSymbol *symbols;
  int numErrors;
  int numWarnings;
  int nextHandle;
  
  struct boundTypeVarEntry {
    ASTTypeVariable *typeVar;
    int scopeHandle;
    struct boundTypeVarEntry *next;
  } *boundTypeVars;
    
  int nextScopeHandle;
  
  TCSymbol *getSymbol(const char *name) { return symbols ? symbols->lookup(name, symbols) : NULL; };
  bool isTypeVariableBound(ASTTypeVariable *typeVar);
  void bindTypeVariableToScope(ASTTypeVariable *typeVar, int handle);
  
public:
  TCContext();
  void addError(ASTExpression *expr, const char *fmt, ...);
  void addWarning(ASTExpression *expr, const char *fmt, ...);
  bool containsSymbol(const char *name) { return getSymbol(name) != NULL; };
  ASTType *getSymbolType(const char *name);
  void addSymbol(const char *name, ASTType *type);
  void removeSymbol(const char *name);
  char *render();
  int getNumErrors() { return numErrors; };
  int getNumWarnings() { return numWarnings; };
  void replaceTypeVarEverywhere(ASTTypeVariable *var, ASTType *replacement);
  int storeType(ASTType *type);
  ASTType *retrieveType(int handle);
  int bindAnyNewTypeVarsIn(ASTType *theType);
  void unbindTypeVarsIn(int scopeHandle);
};

#endif /* defined (__typechecker_h__) */
