#include "global.h"
#include "backend.h"

BECamlVisitor::BECamlVisitor(Printer *printer)
{
  this->printer = printer;
}

BECamlVisitor::~BECamlVisitor()
{
  delete printer;
}

void BECamlVisitor::visit(ASTConstNumber *peer)
{
  printer->print("(INum %.2f)", peer->getValue());
}

void BECamlVisitor::visit(ASTLetBindingExpression *peer)
{
  printer->print("(let %s ", peer->getNameToBind());
  printer->print("= ");
  peer->getRightHandExp()->accept(this);
  printer->print(" in \n");
  peer->getFollowExp()->accept(this);
  printer->print(")");
}

void BECamlVisitor::visit(ASTTypeDefExpression *peer)
{
  peer->getFollowExp()->accept(this); 
}

void BECamlVisitor::visit(ASTLambdaExpression *peer)
{
  const char *fname = peer->getOwnNameOrNull();
  if (fname) {
    printer->print("(let rec %s = IFn (fun %s -> ", fname, peer->getNameToBind());
    printer->indent(+1);
    peer->getInnerExp()->accept(this);
    printer->indent(-1);
    printer->println(") in %s) ", fname);
  }
  else {
    printer->print("(IFn (function %s -> ", peer->getNameToBind());
    printer->indent(+1);
    peer->getInnerExp()->accept(this);
    printer->indent(-1);
    printer->print ("))");  
  }


}

void BECamlVisitor::visit(ASTBaseType *peer)
{
}

void BECamlVisitor::visit(ASTInfixOpExpression *peer)
{
  // binops in the source language mapped uniformly to 
  // curried functions of two args in caml
  const char *op = peer->getOp();
  if (!strcmp(op, "+")) op = "iplus";
  if (!strcmp(op, "-")) op = "iminus";
  if (!strcmp(op, "/")) op = "idiv";
  if (!strcmp(op, "*")) op = "imult";
  if (!strcmp(op, "==")) op = "iequal";
  if (!strcmp(op, ">")) op = "igt";
  if (!strcmp(op, "<")) op = "ilt";
  else if (!strcmp(op, "&&")) op = "iconj";
  else if (!strcmp(op, "||")) op = "idisj";
  printer->print("(%s ", op);
  peer->getLeftExpression()->accept(this);
  printer->print(" ");
  peer->getRightExpression()->accept(this);
  printer->print(")");
}

void BECamlVisitor::visit(ASTVariable *peer)
{
  printer->print("%s", peer->getName());
}

void BECamlVisitor::visit(ASTUnitValue *peer)
{
  printer->print("IUnit");
}

void BECamlVisitor::visit(ASTTypeVariable *peer)
{
}

void BECamlVisitor::visit(ASTNamedType *peer)
{
}

void BECamlVisitor::visit(ASTUnionType *peer)
{
}

void BECamlVisitor::visit(ASTTensorType *peer)
{
}

void BECamlVisitor::visit(ASTAmpersandType *peer)
{
}

void BECamlVisitor::visit(ASTFuzzyType *peer)
{
}

void BECamlVisitor::visit(ASTMuType *peer)
{
}

void BECamlVisitor::visit(ASTCaseExpression *peer)
{
  printer->print("(match ");
  peer->getExpr()->accept(this);
  printer->print(" with ");
  printer->println();
  printer->indent(+1);
  printer->print(" IInl %s -> ", peer->getId(0));
  printer->indent(+1);
  peer->getBranch(0)->accept(this);
  printer->indent(-1);
  printer->print("| IInr %s -> ", peer->getId(1));
  printer->indent(+1);
  peer->getBranch(1)->accept(this);
  printer->indent(-1);
  printer->indent(-1);
  printer->print("| _ -> raise (Failure \"runtime error: noninjection during case\"))"); 
}

void BECamlVisitor::visit(ASTUnfoldExpression *peer)
{
  peer->getInnerExp()->accept(this); 
}

void BECamlVisitor::visit(ASTBuiltinExpression *peer)
{
  printer->print("%s", peer->getName());
}

void BECamlVisitor::visit(ASTFoldExpression *peer)
{
  peer->getInnerExp()->accept(this); 
}

void BECamlVisitor::visit(ASTInjectExpression *peer)
{
  printer->print("(IIn%c ", peer->getBranch() ? 'r' : 'l');
  peer->getInnerExp()->accept(this);
  printer->print(")");
}

void BECamlVisitor::visit(ASTProjectExpression *peer)
{
  printer->print("(%s ", peer->getBranch() ? "isnd" : "ifst");
  peer->getInnerExp()->accept(this);
  printer->print(")");
}

void BECamlVisitor::visit(ASTTensorTuple *peer)
{
  printer->print("(IPair( ");
  peer->getLeftExpression()->accept(this);
  printer->print(", ");
  peer->getRightExpression()->accept(this);
  printer->print("))");
}

void BECamlVisitor::visit(ASTTensorDestructorExpression *peer)
{
  printer->print("(match decons_pair ");
  peer->getProducingExp()->accept(this);
  printer->print(" with (%s, %s) -> ", peer->getName(0), peer->getName(1));
  peer->getConsumingExp()->accept(this); 
  printer->print(")");
}

void BECamlVisitor::visit(ASTAmpersandTuple *peer)
{
  printer->print("(IPair( ");
  peer->getLeftExpression()->accept(this);
  printer->print(", ");
  peer->getRightExpression()->accept(this);
  printer->print("))");
}

void BECamlVisitor::visit(ASTApplyExpression *peer)
{
  printer->print("(iapp ");
  peer->getFun()->accept(this);
  printer->print(" ");
  peer->getArg()->accept(this);
  printer->print(")");
}

void BECamlVisitor::visit(ASTUnitType *peer)
{
}

void BECamlVisitor::visit(ASTFunctionType *peer)
{
}

void BECamlVisitor::visit(ASTSetType *peer)
{
}

void BECamlVisitor::visit(ASTIfThenElseExpression *peer)
{
  printer->print("(if (itrue "); 
  peer->getExpr()->accept(this);
  printer->print(") then ");
  peer->getThenBranch()->accept(this);
  printer->print(" else ");
  peer->getElseBranch()->accept(this);
  printer->print(")");
}

void BECamlVisitor::visit(ASTSampleExpression *peer)
{
  printer->print("(sample (function %s -> ", peer->getNameToBind());
  peer->getFollowExp()->accept(this);
  printer->print(")");
  peer->getRightHandExp()->accept(this);
  printer->print(")");
}

void BECamlVisitor::visit(ASTConstString *peer)
{
  printer->print("(IString \"");
  printer->print(peer->getValue());
  printer->print("\")");
}
