%{
  #include <stdlib.h>
  #include <string.h>
  #include <math.h>

  #include "global.h"
  #include "frontend.h"

  #define GrammarError(x...) do { currentContext->rememberError(linebuf, tokenPos); currentContext->Error(x); } while (0)
  #define currentScope	     (currentContext->getCurrentScope())
  
  extern char *linebuf;
  extern int tokenPos;
  extern ASTExpression *astResult;
  static int enumCounter = 0;

  void yyerror(char *); 
  int yylex();
  
  static SourceContext *getContext()
  
  {
    return new SourceContext(currentContext->filename, linebuf, currentContext->lineno, tokenPos);
  }
%}

%start body

%union {
  const char         *id;
  unsigned int	     integer;
  long double	     real;
  char		     *str;
  char		     chr;
  ASTExpression      *expr;
  ASTType            *type;
  ASTConstExpression *cons;
  FEParameter        *parm;
  SourceContext      *ctxt;
}

%token	<id>		ID
%token	<integer>	LIT_INT
%token	<real>		LIT_REAL
%token	<str>		LIT_STRING
%token	<chr>		LIT_CHR

%token SEMI
%token LBRACE
%token RBRACE
%token COLON
%token COMMA
%token EQUAL
%token ARROW
%token DBLARROW
%token ADD
%token SUB
%token MUL
%token DIV
%token LPAREN
%token RPAREN
%token LT
%token GT
%token LBRACK
%token RBRACK
%token PIPE
%token LOLLIPOP
%token BTRUE
%token BFALSE
%token INF
%token FST
%token SND
%token INL
%token INR
%token FUZZY
%token FUN
%token CASE
%token FOLD
%token UNFOLD
%token MU
%token LET
%token TYPEDEF
%token SAMPLE
%token FUNCTION
%token SET
%token BAG
%token BEQUAL
%token OR
%token AND
%token BANG
%token IF
%token THEN
%token ELSE
%token PRINT

%left MUL DIV MOD ADD SUB ARROW LOLLIPOP
%right FUZZ

%type <id>   id
%type <real> maybe_sens
%type <real> posnum
%type <expr> expr
%type <expr> addex
%type <expr> topex
%type <expr> logex
%type <type> type
%type <type> unarytype
%type <expr> number
%type <expr> mulex
%type <expr> cmpex
%type <expr> fstex
%type <expr> block
%type <expr> term
%type <type> simpletype
%type <type> maybe_typeassert
%type <type> maybe_returntype
%type <id>   maybe_name
%type <parm> parameters
%type <parm> parameter

%%

body            : expr { astResult = $1; }
                
parameter       : LPAREN id COLON maybe_sens type rparen {
                    if (currentScope->symbolExists($2, SYM_LETBINDING))
                      GrammarError("Symbol '%s' redefined", $2);
                    else
                      currentScope->add(SYM_LETBINDING, $2);

                    $$ = new FEParameter($2, $4, $5, getContext());
                  }
                
parameters      : parameters parameter { $$ = $2; $2->append($1); }
                | parameter { $$ = $1; }
                ;
                
expr            : id maybe_typeassert EQUAL topex SEMI {
                    if (currentScope->symbolExists($1, SYM_LETBINDING))
                      GrammarError("Symbol '%s' redefined", $1);
                    else
                      currentScope->add(SYM_LETBINDING, $1);
                  } expr {
                    $$ = new ASTLetBindingExpression($1, $4, $7, $2, getContext());
                  }
                | FUNCTION id {
                    if (currentScope->symbolExists($2, SYM_LETBINDING))
                      GrammarError("Symbol '%s' redefined", $2);
                    else
                      currentScope->add(SYM_LETBINDING, $2);
                    FEScope *newScope = new FEScope(currentScope, getContext());
                    currentContext->setCurrentScope(newScope);
                    $<ctxt>$ = getContext();
                  } parameters COLON type LBRACE expr RBRACE { 
                    currentContext->setCurrentScope((FEScope*)(currentScope->getParentScope()));
                  } expr {
                    FEParameter *iter = $4;
                    ASTExpression *result = $8;
                    ASTType *retType = $6;
                    while (iter) {
                      if (iter->getNext() == NULL) 
                        result = new ASTLambdaExpression(iter->getName(), result, iter->getType(), iter->getSensitivity(), $2, retType, $<ctxt>3);
                      else
                        result = new ASTLambdaExpression(iter->getName(), result, iter->getType(), iter->getSensitivity(), NULL, retType, $<ctxt>3);
                      retType = new ASTFunctionType(iter->getType(), retType, iter->getSensitivity(), $<ctxt>3);
                      iter = iter->getNext();
                    }
                    $$ = new ASTLetBindingExpression($2, result, $11, NULL, $<ctxt>3);
                  }
                | error SEMI expr {
                    GrammarError("Expecting let binding"); $$ = $3; 
                  }
                | SAMPLE id EQUAL topex SEMI {
                    if (currentScope->symbolExists($2))
                      GrammarError("Symbol '%s' redefined", $2);
                    else
                      currentScope->add(SYM_LETBINDING, $2);
                  } expr {
                    $$ = new ASTSampleExpression($2, $4, $7, getContext());
                  }
                | LET LPAREN id {
                    if (currentScope->symbolExists($3))
                      GrammarError("Symbol '%s' redefined", $3);
                  } COMMA id {
                    if (currentScope->symbolExists($6))
                      GrammarError("Symbol '%s' redefined", $6);
                  } RPAREN EQUAL topex SEMI {
                    FEScope *newScope = new FEScope(currentScope, getContext());
                    newScope->add(SYM_LETBINDING, $3);
                    newScope->add(SYM_LETBINDING, $6);
                      
                    currentContext->setCurrentScope(newScope);
                  } expr {
                    $$ = new ASTTensorDestructorExpression($3, $6, $10, $13, getContext());
                    currentContext->setCurrentScope((FEScope*)(currentScope->getParentScope()));
                  }
                | PRINT expr SEMI expr {
                    $$ = new ASTPrintExpression($2, $4, getContext());
                  }
                | TYPEDEF id EQUAL type SEMI {
                    if (currentScope->symbolExists($2))
                      GrammarError("Symbol '%s' redefined", $2);
                    else 
                      currentScope->add(SYM_TYPENAME, $2);
                  } expr {
                    $$ = new ASTTypeDefExpression($2, $4, $7, getContext());
                  }
                | topex 
                ;

maybe_typeassert: { $$ = NULL; }
                | COLON type { $$ = $2; }
                ;

maybe_returntype: { $$ = NULL; }
                | COLON type { $$ = $2; }
                ;

topex           : topex logex {
                    $$ = new ASTApplyExpression($1, $2, getContext());
                  }
                | logex
                ;
    

logex           : logex OR cmpex { $$ = new ASTInfixOpExpression($1, "||", $3, getContext()); }
                | logex AND cmpex { $$ = new ASTInfixOpExpression($1, "&&", $3, getContext()); }
                | BANG cmpex { panic("No bang implemented"); }
                | cmpex
                ;
                
cmpex           : cmpex LT addex { $$ = new ASTInfixOpExpression($1, "<", $3, getContext()); }
                | cmpex GT addex { $$ = new ASTInfixOpExpression($1, ">", $3, getContext()); }
                | cmpex LT EQUAL addex { $$ = new ASTInfixOpExpression($1, "<=", $4, getContext()); }
                | cmpex GT EQUAL addex { $$ = new ASTInfixOpExpression($1, ">=", $4, getContext()); }
                | cmpex BEQUAL addex  { $$ = new ASTInfixOpExpression($1, "==", $3, getContext()); }
                | addex
                ;
                
addex           : addex ADD mulex {
                    if ($1->isCompileTimeConstant() && $3->isCompileTimeConstant()) {
                      $$ = new ASTConstNumber(((ASTConstExpression*)$1)->evaluateAtCompileTime() + ((ASTConstExpression*)$3)->evaluateAtCompileTime(), getContext());
                    } else {
                      $$ = new ASTInfixOpExpression($1, "+", $3, getContext());
                    }
                  }
                | addex SUB mulex {
                    if ($1->isCompileTimeConstant() && $3->isCompileTimeConstant()) {
                      $$ = new ASTConstNumber(((ASTConstExpression*)$1)->evaluateAtCompileTime() - ((ASTConstExpression*)$3)->evaluateAtCompileTime(), getContext());
                    } else {
                      $$ = new ASTInfixOpExpression($1, "-", $3, getContext());
                    }
                  }
                | error ADD mulex { GrammarError("Expecting expression"); $$ = $3; }
                | error SUB mulex { GrammarError("Expecting expression"); $$ = $3; }
                | mulex
                ;
                
mulex           : mulex MUL fstex {
                    if ($1->isCompileTimeConstant() && $3->isCompileTimeConstant()) {
                      $$ = new ASTConstNumber(((ASTConstExpression*)$1)->evaluateAtCompileTime() * ((ASTConstExpression*)$3)->evaluateAtCompileTime(), getContext());
                    } else {
                      $$ = new ASTInfixOpExpression($1, "*", $3, getContext());
                    }
                  }
                | mulex DIV fstex  {
                    if ($1->isCompileTimeConstant() && $3->isCompileTimeConstant()) {
                      if (((ASTConstExpression*)$3)->evaluate() == 0) {
                        GrammarError("Division by zero");
                        $$ = new ASTConstNumber(0, getContext());
                      } else {
                        $$ = new ASTConstNumber(((ASTConstExpression*)$1)->evaluateAtCompileTime() / ((ASTConstExpression*)$3)->evaluateAtCompileTime(), getContext());
                      }
                    } else {
                      $$ = new ASTInfixOpExpression($1, "/", $3, getContext());
                    }
                  }
                | error MUL fstex { GrammarError("Expecting expression"); $$ = $3; }
                | error DIV fstex { GrammarError("Expecting expression"); $$ = $3; }
                | fstex
                ;      
                
fstex           : FST term { $$ = new ASTProjectExpression(0, $2, getContext()); }
                | SND term { $$ = new ASTProjectExpression(1, $2, getContext()); }
                | INL LBRACK type RBRACK term { $$ = new ASTInjectExpression(0, $5, $3, getContext()); }
                | INR LBRACK type RBRACK term { $$ = new ASTInjectExpression(1, $5, $3, getContext()); }
                | FOLD LBRACK type RBRACK fstex { $$ = new ASTFoldExpression($5, $3, getContext()); }
                | UNFOLD term { $$ = new ASTUnfoldExpression($2, getContext()); }
                | term
                ;          

term            : LPAREN RPAREN { $$ = new ASTUnitValue(getContext()); }
                | LPAREN expr RPAREN { $$ = $2; }
                | LPAREN expr COMMA expr RPAREN { $$ = new ASTTensorTuple($2, $4, getContext()); }
                | LPAREN PIPE expr COMMA expr PIPE RPAREN { $$ = new ASTAmpersandTuple($3, $5, getContext()); }
                | CASE expr LBRACE INL LPAREN id {
                    FEScope *newScope = new FEScope(currentScope, getContext());
                    if (!currentScope->symbolExists($6, SYM_LETBINDING))
                      newScope->add(SYM_LETBINDING, $6);
                    else
                      GrammarError("Variable '%s' already exists", $6);
                    currentContext->setCurrentScope(newScope);
                  } RPAREN DBLARROW expr PIPE INR LPAREN id {
                    currentContext->setCurrentScope((FEScope*)(currentScope->getParentScope()));
                    FEScope *newScope = new FEScope(currentScope, getContext());
                    if (!currentScope->symbolExists($14, SYM_LETBINDING))
                      newScope->add(SYM_LETBINDING, $14);
                    else
                      GrammarError("Symbol '%s' redefined", $14);
                    currentContext->setCurrentScope(newScope);
                  } RPAREN DBLARROW expr RBRACE {
                    $$ = new ASTCaseExpression($2, $6, $10, $14, $18, getContext());
                    currentContext->setCurrentScope((FEScope*)(currentScope->getParentScope()));
                  }
                | FUN maybe_name LPAREN id COLON maybe_sens type RPAREN maybe_returntype {
                    FEScope *newScope = new FEScope(currentScope, getContext());
                    if (!currentScope->symbolExists($4))
                      newScope->add(SYM_LETBINDING, $4);
                    else
                      GrammarError("Symbol '%s' redefined", $4);
                    if ($2 && !currentScope->symbolExists($2))
                      newScope->add(SYM_LETBINDING, $2);
                    else if ($2)
                      GrammarError("Symbol '%s' redefined", $2);
                    currentContext->setCurrentScope(newScope);
                  } block {
                    $$ = new ASTLambdaExpression($4, $11, $7, $6, $2, $9, getContext());
                    currentContext->setCurrentScope((FEScope*)(currentScope->getParentScope()));
                    if ($2 && !$9) {
                      GrammarError("Named function '%s' needs to be given a return type", $2);
                      $$ = new ASTConstNumber(0, getContext());
                    }
                  }
                | IF expr THEN LBRACE expr RBRACE ELSE LBRACE expr RBRACE {
                    $$ = new ASTIfThenElseExpression($2, $5, $9, getContext());
                  }
                | ID {
                    if (currentScope->symbolExists($1, SYM_LETBINDING)) {
                      $$ = new ASTVariable($1, getContext());
                    } else {
                      GrammarError("Unknown identifier: '%s'", $1);
                      $$ = new ASTConstNumber(0, getContext());
                    }
                  }
                | number
                | LIT_STRING { $$ = new ASTConstString($1); }
                | BTRUE { $$ = new ASTConstBool(true, getContext()); }
                | BFALSE { $$ = new ASTConstBool(false, getContext()); }
                ;

block           : LBRACE expr RBRACE { $$ = $2; }
                | LBRACE error RBRACE { GrammarError("Expecting block"); $$ = new ASTConstNumber(0, getContext()); }
                ;

maybe_name      : { $$ = NULL; }
                | ID { $$ = $1; }
                ;

type            : type ADD unarytype { $$ = new ASTUnionType($1, $3, getContext()); }
                | type LOLLIPOP maybe_sens unarytype { if ($3==INFINITY) $3=1; $$ = new ASTFunctionType($1, $4, $3, getContext()); }
                | type ARROW unarytype { $$ = new ASTFunctionType($1, $3, INFINITY, getContext()); };
                | type SET {
                    $$ = new ASTSetType($1, false, getContext());
                  }
                | type BAG {
                    $$ = new ASTSetType($1, true, getContext());
                  }
                | MU id {
                    FEScope *newScope = new FEScope(currentScope, getContext());
                    if (!currentScope->symbolExists($2))
                      newScope->add(SYM_TYPEVAR, $2);
                    else
                      GrammarError("Symbol '%s' redefined", $2);
                    currentContext->setCurrentScope(newScope);
                  } DBLARROW unarytype { 
                    $$ = new ASTMuType($2, $5, getContext());
                    currentContext->setCurrentScope((FEScope*)(currentScope->getParentScope()));
                  }
                | unarytype
                ;
                
unarytype       : FUZZY simpletype { $$ = new ASTFuzzyType($2, getContext()); }
                | simpletype
                ;
                
simpletype      : LPAREN RPAREN { $$ = new ASTUnitType(getContext()); };
                | ID {
                    if (!strcmp($1, "num") || !strcmp($1, "bool") || !strcmp($1, "clipped") || !strcmp($1, "string")) {
                      $$ = new ASTBaseType($1, getContext());
                    } else if (currentScope->symbolExists($1, SYM_TYPENAME)) {
                      $$ = new ASTNamedType($1, getContext());
                    } else {
                      if (!currentScope->symbolExists($1, SYM_TYPEVAR)) 
                        currentScope->add(SYM_TYPEVAR, $1);
                        
                      $$ = new ASTTypeVariable($1, 0, getContext());
                    }
                  }
                | LPAREN type COMMA type rparen { $$ = new ASTTensorType($2, $4, getContext()); }
                | LPAREN PIPE type COMMA type PIPE rparen { $$ = new ASTAmpersandType($3, $5, getContext()); }
                | LPAREN type rparen { $$ = $2; }
                ;

posnum          : LIT_INT { 
                    if ($1 <= 0) {
                      GrammarError("Expecting sensitivity (must be strictly positive!)");
                      $$ = 1;
                    } else {
                      $$ = $1;
                    }
                  }
                | LIT_REAL { 
                    if ($1 <= 0) {
                      GrammarError("Expecting sensitivity (must be strictly positive!)");
                      $$ = 1;
                    } else {
                      $$ = $1;
                    }
                  }
                | error { GrammarError("Expecting positive number"); $$ = 1; }
                ;

maybe_sens      : { $$ = INFINITY; }
                | LBRACK posnum RBRACK { $$ = $2; }
                | LBRACK RBRACK { $$ = 1; }
                ;

number          : LIT_INT { $$ = new ASTConstNumber($1, getContext()); }
                | LIT_REAL { $$ = new ASTConstNumber($1, getContext()); }
                ;

id		: ID
		;
                
rparen		: RPAREN
		| error { GrammarError("Expecting )"); }
		;

%%

void
yyerror(char * s)
{
  GrammarError(s);
}
