#include <assert.h>

#include "global.h"
#include "ast.h"
#include "printer.h"
#include "dump.h"

#define BEGIN_EVAL(exp) do { \
  extern bool debugMode; \
  if (debugMode) { \
    if (!printer) \
      printer = new Printer(stdout); \
    printer->println("Eval: [%s]", render(exp)); \
    printer->indent(+1); \
  } \
} while (0)
#define END_EVAL(exp, res) do { \
  extern bool debugMode; \
  if (debugMode) { \
    char buf[2000]; \
    printer->indent(-1); \
    printer->println("Res: %s", render(res)); \
  } \
} while (0)

static Printer *printer = NULL;

/* 
 * Convention: applyTo can consume 'exp', and the caller can consume the return value.
 * If you're planning to return some part of the AST tree, you must clone it!
*/

ASTExpression *ASTConstNumber::evaluate()
{
  BEGIN_EVAL(this);
  
  ASTExpression *result = new ASTConstNumber(this->value);
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTConstBool::evaluate()
{
  BEGIN_EVAL(this);
  
  ASTExpression *result = new ASTConstBool(this->value);
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTConstString::evaluate()
{
  BEGIN_EVAL(this);
  
  ASTExpression *result = new ASTConstString(this->value);
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTVariable::evaluate()
{
  assert(false);
}

ASTExpression *ASTInfixOpExpression::evaluate()
{
  BEGIN_EVAL(this);
  
  ASTExpression *leftResult = left->evaluate();
  ASTExpression *rightResult = right->evaluate();
  assert(leftResult && rightResult);
  
  ASTExpression *result;
  
  if (!strcmp(op, "+") || !strcmp(op, "*") || !strcmp(op, "/") || !strcmp(op, "-")) {
    if (!strcmp(op, "+") && (leftResult->getExpCode() == EXP_CONSTSTRING) && (rightResult->getExpCode() == EXP_CONSTSTRING)) {
      const char *l = ((ASTConstString*)leftResult)->getValue();
      const char *r = ((ASTConstString*)rightResult)->getValue();
      char *res = (char*)malloc(strlen(l)+strlen(r)+1);
      strcpy(res, l);
      strcat(res, r);
      result = new ASTConstString(res);
      free(res);
    } else if ((leftResult->getExpCode() == EXP_CONSTNUM) && (rightResult->getExpCode() == EXP_CONSTNUM)) {
      double l = ((ASTConstNumber*)leftResult)->getValue();
      double r = ((ASTConstNumber*)rightResult)->getValue();
      double res = 0;
      if (!strcmp(op, "+")) {
        res = l+r;
      } else if (!strcmp(op, "-")) {
        res = l-r;
      } else if (!strcmp(op, "*")) {
        res = l*r;
      } else if (!strcmp(op, "/")) {
        res = l/r;
      } else {
        panic("Cannot evaluate infix operator: '%s'", op);
      }

      delete leftResult;
      delete rightResult;
      result = new ASTConstNumber(res);
    } else {
      result = new ASTInfixOpExpression(leftResult, op, rightResult);
    }
  } else if (!strcmp(op, "==") || !strcmp(op, "<") || !strcmp(op, ">") || !strcmp(op, "<=") || !strcmp(op, ">=")) {
    assert(leftResult->getExpCode() == EXP_CONSTNUM);
    assert(rightResult->getExpCode() == EXP_CONSTNUM);
    double l = ((ASTConstNumber*)leftResult)->getValue();
    double r = ((ASTConstNumber*)rightResult)->getValue();
    bool res;
    if (!strcmp(op, "=="))
      res = (l==r); 
    else if (!strcmp(op, "<"))
      res = (l<r);
    else if (!strcmp(op, ">"))
      res = (l>r);
    else if (!strcmp(op, "<="))
      res = (l<=r);
    else if (!strcmp(op, ">="))
      res = (l>=r);
      
    result = new ASTConstBool(res);
  } else if (!strcmp(op, "&&") || !strcmp(op, "||")) {
    assert(leftResult->getExpCode() == EXP_CONSTBOOL);
    assert(rightResult->getExpCode() == EXP_CONSTBOOL);
    bool l = ((ASTConstBool*)leftResult)->getValue();
    bool r = ((ASTConstBool*)rightResult)->getValue();
    bool res;
    if (!strcmp(op, "&&"))
      res = (l && r); 
    else if (!strcmp(op, "||"))
      res = (l || r);
      
    result = new ASTConstBool(res);
  } else {
    panic("Evaluation not implemented for infix operator '%s'", op);
  }
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTUnfoldExpression::evaluate()
{
  BEGIN_EVAL(this);

  ASTExpression *result = innerExp->evaluate();
  assert (result->getExpCode() == EXP_FOLD);

  ASTFoldExpression *foldex = (ASTFoldExpression*) result;
  result = foldex->getInnerExp();
    
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTBuiltinExpression::applyTo(ASTExpression *arg)
{
  if (!strcasecmp(name, "return")) {
    assert(arg);
    return new ASTSingleValueDistribution(arg);
  } else if (!strcasecmp(name, "fuzz")) {
    assert(arg && (arg->getExpCode() == EXP_CONSTNUM));
    return new ASTLaplaceDistribution(((ASTConstNumber*)arg)->getValue());
  } else if (!strcasecmp(name, "count")) {
    assert(arg && (arg->getExpCode() == EXP_SET));
    return new ASTConstNumber(((ASTSetExpression*)arg)->count());
  } else if (!strcasecmp(name, "clip")) {
    assert(arg && (arg->getExpCode() == EXP_CONSTNUM));
    double val = ((ASTConstNumber*)arg)->getValue();
    return new ASTConstNumber((val<-1) ? -1 : ((val>1) ? 1 : val));
  } else if (!strcasecmp(name, "bagsize")) {
    assert(arg->getExpCode() == EXP_BAG);
    ASTSetExpression *argAsSet = (ASTSetExpression*) arg;
    return new ASTConstNumber(argAsSet->count());
  } else if (!strcasecmp(name, "random")) {
    if (numArgs < 1) {
      ASTBuiltinExpression *retval = clone();
      retval->addArgument(arg->clone());
      return retval;
    }
    
    assert(arg->getExpCode() == EXP_CONSTNUM);
    assert(myArg[0]->getExpCode() == EXP_CONSTNUM);
    double lower = ((ASTConstNumber*)(myArg[0]))->getValue();
    double upper = ((ASTConstNumber*)arg)->getValue();
    double r = lower + ((disableRandom ? ((random()%1000000)*0.000001) : drand48())*(upper-lower));
    return new ASTConstNumber(r);
    
  } else if (!strcasecmp(name, "setsum") || !strcasecmp(name, "bagsum")) {
    bool bagMode = !strcasecmp(name, "bagsum");
    assert(arg->getExpCode() == (bagMode ? EXP_BAG : EXP_SET));
    
    ASTSetExpression *argAsSet = (ASTSetExpression*) arg;
    struct ASTSetExpression::elementStruct *iter = argAsSet->elements;
    double resultVal = 0;
    while (iter) { 
      assert(iter->value->getExpCode() == EXP_CONSTNUM);
      resultVal += ((ASTConstNumber*)(iter->value))->getValue();
      iter = iter->next;
    }
    
    return new ASTConstNumber(resultVal);
  } else if (!strcasecmp(name, "setmap") || !strcasecmp(name, "bagmap")) {
#ifdef USE_TIMEOUTS_AND_DEFAULTS
    if (numArgs < 3) {
      ASTBuiltinExpression *retval = clone();
      retval->addArgument(arg->clone());
      return retval;
    }
    assert(myArg[2]->getExpCode() == EXP_CONSTNUM);
#else
    if (numArgs < 1) {
      ASTBuiltinExpression *retval = clone();
      retval->addArgument(arg->clone());
      return retval;
    }
#endif
      
    bool bagMode = !strcasecmp(name, "bagmap");
    assert(arg->getExpCode() == (bagMode ? EXP_BAG : EXP_SET));
    assert(myArg[0]->getExpCode() == EXP_LAMBDA);
    
    ASTSetExpression *argAsSet = (ASTSetExpression*) arg;
    ASTLambdaExpression *myargAsLambda = (ASTLambdaExpression*) myArg[0];
    
    ASTType *returnType = myargAsLambda->getReturnTypeOrNull();
    if (!returnType)
      panic("%s is being applied to a function (%s) without a declared return type", name, render(myArg[0]));
      
    /* Here, we are ignoring the timeout and the default value (the real backend uses them) */
      
    ASTSetExpression *result = new ASTSetExpression(returnType->clone(), bagMode);
    struct ASTSetExpression::elementStruct *iter = argAsSet->elements;
    while (iter) { 
      result->add(myArg[0]->applyTo(iter->value));
      iter = iter->next;
    }
    
    return result;
  } else if (!strcasecmp(name, "bagsplit") || !strcasecmp(name, "setsplit")) {
#ifdef USE_TIMEOUTS_AND_DEFAULTS
    if (numArgs < 2) {
      ASTBuiltinExpression *retval = clone();
      retval->addArgument(arg->clone());
      return retval;
    }
    assert(myArg[1]->getExpCode() == EXP_CONSTNUM);
#else
    if (numArgs < 1) {
      ASTBuiltinExpression *retval = clone();
      retval->addArgument(arg->clone());
      return retval;
    }
#endif
    
    bool bagMode = !strcasecmp(name, "bagsplit");
    assert(arg->getExpCode() == (bagMode ? EXP_BAG : EXP_SET));
    assert(myArg[0]->getExpCode() == EXP_LAMBDA);
    
    ASTSetExpression *argAsSetOrBag = (ASTSetExpression*) arg;
    ASTLambdaExpression *myargAsLambda = (ASTLambdaExpression*) myArg[0];
    
    /* Here, we are ignoring the timeout and the default value (the real backend uses them) */
      
    ASTSetExpression *is = new ASTSetExpression(argAsSetOrBag->getMemberType(), bagMode);
    ASTSetExpression *isNot = new ASTSetExpression(argAsSetOrBag->getMemberType(), bagMode);
    struct ASTSetExpression::elementStruct *iter = argAsSetOrBag->elements;
    while (iter) { 
      ASTExpression *res = myArg[0]->applyTo(iter->value);
      assert(res && res->getExpCode() == EXP_CONSTBOOL);
      ASTConstBool *resAsBool = (ASTConstBool*) res;
      if (resAsBool->getValue()) 
        is->add(iter->value);
      else
        isNot->add(iter->value);
      iter = iter->next;
    }
    
    return new ASTTensorTuple(is, isNot);
  } else if (!strcasecmp(name, "cswp")) {
    assert(arg->getExpCode() == EXP_TENSOR);
    ASTTensorTuple *tens = (ASTTensorTuple*) arg;
    assert(tens->getLeftExpression()->getExpCode() == EXP_CONSTNUM);
    assert(tens->getRightExpression()->getExpCode() == EXP_CONSTNUM);
    
    ASTConstNumber *a = (ASTConstNumber*) tens->getLeftExpression();
    ASTConstNumber *b = (ASTConstNumber*) tens->getRightExpression();
    if (a->getValue() < b->getValue())
      return new ASTTensorTuple(a->clone(), b->clone());
    else
      return new ASTTensorTuple(b->clone(), a->clone());
  } else if (!strcasecmp(name, "num2string")) {
    assert(arg->getExpCode() == EXP_CONSTNUM);
    ASTConstNumber *cn = (ASTConstNumber*) arg;
    char s[500];
    prettyDouble(cn->getValue(), s);
    return new ASTConstString(s);
  } else {
    panic("Not implemented: applyTo (ASTBuiltinExpression %s)", name);
  }
}

ASTExpression *ASTBuiltinExpression::evaluate()
{
  BEGIN_EVAL(this);

  ASTExpression *result = clone();  
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTFoldExpression::evaluate()
{
  BEGIN_EVAL(this);

  ASTExpression *result = new ASTFoldExpression(innerExp->evaluate(), typeArg->clone());
    
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTProjectExpression::evaluate()
{
  BEGIN_EVAL(this);

  assert(innerExp->getExpCode() == EXP_AMPERSAND);
  ASTAmpersandTuple *tuple = (ASTAmpersandTuple*)innerExp;
  ASTExpression *theExp = branch ? tuple->getRightExpression() : tuple->getLeftExpression();
  ASTExpression *result = theExp->evaluate();
    
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTCaseExpression::evaluate()
{
  BEGIN_EVAL(this);

  ASTExpression *exprEval = expr->evaluate();
  assert(exprEval->getExpCode() == EXP_INJ);

  ASTInjectExpression *inj = (ASTInjectExpression*) exprEval;
  int whichBranch = inj->getBranch();

  ASTExpression *result = branch[whichBranch]->substitute(id[whichBranch], inj->getInnerExp())->evaluate();
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTIfThenElseExpression::evaluate()
{
  BEGIN_EVAL(this);

  ASTExpression *exprEval = expr->evaluate();
  assert(exprEval->getExpCode() == EXP_CONSTBOOL);

  ASTExpression *result = (((ASTConstBool*) exprEval)->getValue()) ? thenBranch->evaluate() : elseBranch->evaluate();
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTInjectExpression::evaluate()
{
  BEGIN_EVAL(this);

  ASTExpression *innerEval = innerExp->evaluate();
  ASTExpression *result = new ASTInjectExpression(branch, innerEval, typeArg->clone());
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTTensorTuple::evaluate()
{
  BEGIN_EVAL(this);

  ASTExpression *leftEval = left->evaluate();
  ASTExpression *rightEval = right->evaluate();
  ASTExpression *result = new ASTTensorTuple(leftEval, rightEval);
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTTensorDestructorExpression::evaluate()
{
  BEGIN_EVAL(this);
  
  ASTExpression *theTensor = producingExp->evaluate();
  assert(theTensor->getExpCode() == EXP_TENSOR);

  ASTTensorTuple *tuple = (ASTTensorTuple*)theTensor;
  ASTExpression *afterSubst1 = consumingExp->substitute(name1, tuple->getLeftExpression());
  ASTExpression *afterSubst2 = afterSubst1->substitute(name2, tuple->getRightExpression());
  ASTExpression *result = afterSubst2->evaluate();
   
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTAmpersandTuple::evaluate()
{
  BEGIN_EVAL(this);

  ASTExpression *leftEval = left->evaluate();
  ASTExpression *rightEval = right->evaluate();
  ASTExpression *result = new ASTAmpersandTuple(leftEval, rightEval);
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTLambdaExpression::evaluate()
{
  BEGIN_EVAL(this);

  ASTExpression *result = clone();
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTLambdaExpression::applyTo(ASTExpression *arg)
{
  ASTExpression *selfSubst = ownNameOrNull ? innerExp->substitute(ownNameOrNull, this) : innerExp->clone();
  ASTExpression *result = selfSubst->substitute(nameToBind, arg)->evaluate();
  return result;
}

ASTExpression *ASTApplyExpression::evaluate()
{
  BEGIN_EVAL(this);

  ASTExpression *funEval = fun->evaluate();
  ASTExpression *argEval = arg->evaluate();
  ASTExpression *result = funEval->applyTo(argEval);
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTUnitValue::evaluate()
{
  BEGIN_EVAL(this);
  
  ASTExpression *result = new ASTUnitValue();
  
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTLetBindingExpression::evaluate()
{
  BEGIN_EVAL(this);
  
  ASTExpression *rheEval = rightHandExp->evaluate();
  ASTExpression *afterSubst = followExp->substitute(nameToBind, rheEval);
  ASTExpression *result = afterSubst->evaluate();
   
  END_EVAL(exp, result);
  return result;
}

extern bool enablePrintExp;

ASTExpression *ASTPrintExpression::evaluate()
{
  BEGIN_EVAL(this);
  
  ASTExpression *rheEval = rightHandExp->evaluate();
  if (enablePrintExp) {
    if (!printer) 
      printer = new Printer(stdout);
    printer->println("%s", render(rheEval));
  }
    
  ASTExpression *result = followExp->evaluate();
   
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTTypeDefExpression::evaluate()
{
  BEGIN_EVAL(this);
  
  ASTExpression *result = followExp->evaluate();
    
  END_EVAL(exp, result);
  return result;
}

ASTExpression *ASTSampleExpression::evaluate()
{
  BEGIN_EVAL(this);

  ASTExpression *inner = rightHandExp->evaluate();
  assert(inner && (inner->getExpCode() == EXP_DISTRIBUTION));

  ASTDistribution *innerDist = (ASTDistribution*) inner;
  ASTExpression *sampleVal = innerDist->sample();

  ASTExpression *afterSubst = followExp->substitute(nameToBind, sampleVal);
  ASTExpression *result = afterSubst->evaluate();
 
  END_EVAL(exp, result);
  return result;
}

double laplaceDraw(double mu, double b)
{
  double r = drand48();
  if (r < 0.5)
    return mu + b*log(2*r);
  else 
    return mu - b*log(2*(1-r));
}
