#include <stdlib.h>
#include <fcntl.h>

#include "global.h"
#include "frontend.h"
#include "ast.h"
#include "typechecker.h"
#include "backend.h"
#include "dump.h"

bool debugMode = false;
bool verboseMode = false;
bool disableRandom = false;
bool disableFuzz = false;
bool enablePrettyPrinting = true;
bool enablePrintExp = false;
double epsilon = 0.01;
int enabledWarnings = 0;

ASTExpression *readDataFile(const char *datafileName, const char *dataFormat)
{
  FILE *infile = !strcmp(datafileName, "-") ? stdin : fopen(datafileName, "r");
  if (!infile) {
    fprintf(stderr, "Cannot open input file: '%s'\n", datafileName);
    exit(1);
  }

  ASTExpression *result = NULL;
  char linebuf[2000];
  
  if (!strcasecmp(dataFormat, "num")) {
    fgets(linebuf, sizeof(linebuf), infile);
    strtok(linebuf, "#\r\n");
    result = new ASTConstNumber(atof(linebuf), new SourceContext(datafileName, linebuf, 1, 1));
  } else if (!strcasecmp(dataFormat, "numlist")) {
    ASTType *listType = new ASTMuType("X", new ASTUnionType(new ASTUnitType(), new ASTTensorType(new ASTBaseType("num"), new ASTTypeVariable("X"))));
    ASTType *innerType = new ASTUnionType(new ASTUnitType(), new ASTTensorType(new ASTBaseType("num"), listType->clone()));

    ASTTensorTuple *tens = NULL;
    while (fgets(linebuf, sizeof(linebuf), infile)) {
      ASTTensorTuple *newTens = new ASTTensorTuple(new ASTConstNumber(atof(linebuf), new SourceContext(datafileName, linebuf, 1, 1)), NULL);
      ASTFoldExpression *newFold = new ASTFoldExpression(new ASTInjectExpression(1, newTens, innerType->clone()), listType->clone());
      if (!result)
        result = newFold;
      else
        tens->setRightExpression(newFold);
        
      tens = newTens;
    }

    ASTExpression *terminal = new ASTFoldExpression(new ASTInjectExpression(0, new ASTUnitValue(), innerType->clone()), listType->clone());
    if (!result)
      result = terminal;
    else
      tens->setRightExpression(terminal);
  } else if (!strcasecmp(dataFormat, "numset") || !strcasecmp(dataFormat, "numbag")) {
    int lineno = 1;
    fgets(linebuf, sizeof(linebuf), infile);
    char *arg[20];
    int cols = 1;
    arg[0] = strtok(linebuf, " \r\n");
    for (int i=1; i<20; i++) {
      arg[i] = arg[i-1] ? strtok(NULL, " \r\n") : NULL;
      if (arg[i])
        cols ++;
    }
    
    ASTType *elementType = new ASTBaseType("num");
    for (int i=1; i<cols; i++)
      elementType = new ASTTensorType(new ASTBaseType("num"), elementType);
    
    ASTSetExpression *theSet = new ASTSetExpression(elementType, !strcasecmp(dataFormat, "numbag"), new SourceContext(datafileName, "", 1, 1));
    do {
      ASTExpression *element = new ASTConstNumber(atof(arg[cols-1]));
      for (int i=cols-2; i>=0; i--)
        element = new ASTTensorTuple(new ASTConstNumber(atof(arg[i])), element);
        
      theSet->add(element);
      
      if (!fgets(linebuf, sizeof(linebuf), infile))
        break;
        
      lineno ++;
      arg[0] = strtok(linebuf, " \r\n");
      for (int i=1; i<20; i++) 
        arg[i] = arg[i-1] ? strtok(NULL, " \r\n") : NULL;
      
      int colsHere = 0;
      while ((colsHere<20) && arg[colsHere])
        colsHere ++;
      
      if (colsHere != cols)
        panic("Number of columns in data file '%s' is inconsistent -- first line has %d, but line %d has %d", datafileName, cols, lineno, colsHere);
    } while (true);
    
    result = theSet;
  } else {
    panic("Unknown data file format: '%s'", dataFormat);
  }
  
  fclose(infile);
  
  return result;
}

char *render(ASTBase *node, bool singleOnly)
{
  int bufsize = 800;
  while (true) {
    char *buf = (char*) malloc(bufsize);
    Printer *stringPrinter = new Printer(buf, bufsize);
    ASTDumpVisitor *dumper = new ASTDumpVisitor(stringPrinter, singleOnly);
  
    stringPrinter->setLineMode(true);
    node->accept(dumper);
    stringPrinter->flush();
    bool full = stringPrinter->bufferFull();
    
    delete dumper;
    
    if (!full)
      return buf;
    
    strcpy(&buf[bufsize-6], " ...");
    return buf;
    
/*    free(buf);
    bufsize *= 2;
    assert(bufsize < 1048576); */
  }
}

int main(int argc, char *argv[])
{
  const char *codefileName = NULL;
  const char *datafileName = NULL;
  const char *outfileName = NULL;
  const char *dataFormat = "numlist";
  bool helpRequested = false;
  bool allowUnsampled = false;
  bool useStrongRandomness = false;
  char *cppOptions = (char*) malloc(1);
  *cppOptions = 0;
  
  for (int argi=1; argi<argc; argi++) {
    if (!strcmp(argv[argi], "--help")) {
      helpRequested = true;
      
    // Verbose mode: Prints extra information during type checking etc.
      
    } else if (!strcmp(argv[argi], "--verbose")) {
      verboseMode = true;
      
    // Permissive mode: Allows user to run programs with non-monadic types
      
    } else if (!strcmp(argv[argi], "--permissive")) {
      allowUnsampled = true;
      
    // Print some extra warnings for debugging
      
    } else if (!strncmp(argv[argi], "-W", 2)) {
      char *ptr = strtok(&argv[argi][2], ",");
      while (ptr) {
        if (!strcmp(ptr, "tight"))
          enabledWarnings |= WARN_TIGHT_SENSITIVITIES;
        else
          panic("Unknown warning: '%s'", ptr);
        ptr = strtok(NULL, ",");
      }
      
    // Output file name (this is the file that will later be compiled with Caml)
      
    } else if (!strcmp(argv[argi], "-o")) {
      outfileName = argv[++argi];
      
    // Evaluation mode: Uses the built-in interpreter rather than produce Caml code [for testing - do not use!]
      
    } else if (!strcmp(argv[argi], "--eval")) {
      datafileName = argv[++argi];
      
    // Use /dev/urandom to 'properly' initialize random number generator (may be slower).
    // If you don't use this, the adversary may be able to predict your random seed.
      
    } else if (!strcmp(argv[argi], "--use-strong-randomness")) {
      useStrongRandomness = true;
      
    // For use with --eval (sets the data format of the input file)
      
    } else if (!strcmp(argv[argi], "--format")) {
      dataFormat = argv[++argi];
      
    // Obsolete
      
    } else if (!strcmp(argv[argi], "--disable-fuzz")) {
      disableFuzz = true;
      
    // Don't fuzz the output (used for the test suite)
      
    } else if (!strcmp(argv[argi], "--disable-random")) {
      disableRandom = true;
      
    // Allow printing to the terminal (useful for debugging, but certainly not when running on 'real' data)
      
    } else if (!strcmp(argv[argi], "--enable-print")) {
      enablePrintExp = true;
      
    // Do not do pretty-printing
      
    } else if (!strcmp(argv[argi], "--nopretty")) {
      enablePrettyPrinting = false;
      
    // Specify epsilon
      
    } else if (!strcmp(argv[argi], "--epsilon")) {
      epsilon = atof(argv[++argi]);
      if ((epsilon <= 0) || (epsilon >= 1))
        panic("Epsilon must be between zero and one!");
        
    // Specify include directory for cpp (so we can have macros and hash-include things)
        
    } else if (!strcmp(argv[argi], "-I")) {
      char *newbuf = (char*) malloc(strlen(cppOptions) + 4 + strlen(argv[argi+1]));
      sprintf(newbuf, "%s -I %s", cppOptions, argv[++argi]);
      free(cppOptions);
      cppOptions = newbuf;
      
    // print version number and exit  
      
    } else if (!strcmp(argv[argi], "--version")) {
      printf("%s\n", VERSION);
      return 1;
      
    // enable debug mode
      
    } else if (!strcmp(argv[argi], "--debug")) {
      debugMode = true;
    } else if (argv[argi][0]=='-' && argv[argi][1]) {
      fprintf(stderr, "Unknown option: '%s'\n", argv[argi]);
      return 1;
    } else {
      if (!codefileName) {
        codefileName = argv[argi];
      } else {
        fprintf(stderr, "Extra argument: '%s'\n", argv[argi]);
        return 1;
      }
    }
  }

  if (!codefileName || helpRequested || (outfileName && datafileName)) {
    fprintf(stderr, "Syntax: fuzz [--verbose] [--permissive] [-I <includedir>] [--epsilon <e>] [--format <dataformat>] <codefile> [--eval <datafile>] [-o <outfile>]\n");
    return 1;
  }

  /* We need really good, unpredictable randomness, which we get from /dev/random or /dev/urandom
     (were we merely to call srand(time(0)), an adversary might be able to reverse-engineer our
     randomness if they knew the approxmimate time at which the function was evaluated). The
     difference is that with /dev/random/, we may have to block for a few seconds if the randomness
     bucket happens to be empty. */
  
  const char *devRandom = useStrongRandomness ? "/dev/random" : "/dev/urandom";
  unsigned char seed16v[6];
  int h = open(devRandom, O_RDONLY);
  if (h < 0)
    panic("Cannot open %s", devRandom);
  int r = read(h, seed16v, sizeof(seed16v));
  if (r != sizeof(seed16v))
    panic("Cannot read from %s", devRandom);
  close(h);
  seed48((unsigned short*)&seed16v);

  /* Parse */
  
  FEScope *rootScope = new FEScope(NULL, new SourceContext());
  rootScope->add(SYM_LETBINDING, "fuzz");
  rootScope->add(SYM_LETBINDING, "return");
  rootScope->add(SYM_LETBINDING, "bagsize");
  rootScope->add(SYM_LETBINDING, "clip");
  rootScope->add(SYM_LETBINDING, "bagsplit");
  rootScope->add(SYM_LETBINDING, "cswp");
  rootScope->add(SYM_LETBINDING, "setsum");
  rootScope->add(SYM_LETBINDING, "bagsum");
  rootScope->add(SYM_LETBINDING, "setmap");
  rootScope->add(SYM_LETBINDING, "bagmap");
  rootScope->add(SYM_LETBINDING, "setsplit");
  rootScope->add(SYM_LETBINDING, "random");
  rootScope->add(SYM_LETBINDING, "num2string");

  InputContext *frontend = new InputContext(codefileName);
  frontend->setOptions(cppOptions);
  
  ASTExpression *ast = frontend->parse(rootScope);
  
  if (frontend->hasErrors()) {
    fprintf(stderr, "%s: %i Error(s) and %i Warning(s)\n", codefileName,
      frontend->getErrorCount(), frontend->getWarningCount());
    return 1;
  }

  if (verboseMode) {
    printf("=== Input from %s:\n", codefileName);
    Printer *prn = new Printer(stdout);
    ASTDumpVisitor *astDumper = new ASTDumpVisitor(prn);
    ast->accept(astDumper);
    delete astDumper;
  }

  /* Type check */

  if (verboseMode)
    printf("\n=== Type inference:\n");

  TCContext *ctx = new TCContext();
  ctx->addSymbol("fuzz", new ASTFunctionType(new ASTBaseType("num"), new ASTFuzzyType(new ASTBaseType("num")), 1));
  ctx->addSymbol("return", new ASTFunctionType(new ASTTypeVariable("_return"), new ASTFuzzyType(new ASTTypeVariable("_return")), 1));
  ctx->addSymbol("bagsize", new ASTFunctionType(new ASTSetType(new ASTTypeVariable("_bagsize"), true), new ASTBaseType("num"), 1));
  ctx->addSymbol("clip", new ASTFunctionType(new ASTBaseType("num"), new ASTBaseType("clipped"), 1));
#ifdef USE_TIMEOUTS_AND_DEFAULTS
  ctx->addSymbol("bagsplit", 
    new ASTFunctionType(
      new ASTFunctionType(new ASTTypeVariable("_bagsplit"), new ASTBaseType("bool"), INFINITY), 
      new ASTFunctionType(
        new ASTBaseType("num"),
        new ASTFunctionType(
          new ASTSetType(new ASTTypeVariable("_bagsplit"), true), 
          new ASTTensorType(
            new ASTSetType(new ASTTypeVariable("_bagsplit"), true),
            new ASTSetType(new ASTTypeVariable("_bagsplit"), true)
        ), 1), 
      INFINITY),
    INFINITY)
  );
  ctx->addSymbol("setsplit", 
    new ASTFunctionType(
      new ASTFunctionType(new ASTTypeVariable("_setsplit"), new ASTBaseType("bool"), INFINITY), 
      new ASTFunctionType(
        new ASTBaseType("num"),
        new ASTFunctionType(
          new ASTSetType(new ASTTypeVariable("_setsplit"), false), 
          new ASTTensorType(
            new ASTSetType(new ASTTypeVariable("_setsplit"), false),
            new ASTSetType(new ASTTypeVariable("_setsplit"), false)
        ), 1), 
      INFINITY),
    INFINITY)
  );
  ctx->addSymbol("setmap", 
    new ASTFunctionType(
      new ASTFunctionType(new ASTTypeVariable("_setmap1"), new ASTTypeVariable("_setmap2"), INFINITY), 
      new ASTFunctionType(
        new ASTTypeVariable("_setmap2"),
        new ASTFunctionType(
          new ASTBaseType("num"),
          new ASTFunctionType(
            new ASTSetType(new ASTTypeVariable("_setmap1"), false), 
            new ASTSetType(new ASTTypeVariable("_setmap2"), false), 
          1), 
        INFINITY),
      INFINITY),
    INFINITY)
  );
  ctx->addSymbol("bagmap", 
    new ASTFunctionType(
      new ASTFunctionType(new ASTTypeVariable("_bagmap1"), new ASTTypeVariable("_bagmap2"), INFINITY), 
      new ASTFunctionType(
        new ASTTypeVariable("_bagmap2"),
        new ASTFunctionType(
          new ASTBaseType("num"),
          new ASTFunctionType(
            new ASTSetType(new ASTTypeVariable("_bagmap1"), true), 
            new ASTSetType(new ASTTypeVariable("_bagmap2"), true), 
          1), 
        INFINITY),
      INFINITY),
    INFINITY)
  );
#else
  ctx->addSymbol("bagsplit", 
    new ASTFunctionType(
      new ASTFunctionType(new ASTTypeVariable("_bagsplit"), new ASTBaseType("bool"), INFINITY), 
      new ASTFunctionType(
        new ASTSetType(new ASTTypeVariable("_bagsplit"), true), 
        new ASTTensorType(
          new ASTSetType(new ASTTypeVariable("_bagsplit"), true),
          new ASTSetType(new ASTTypeVariable("_bagsplit"), true)
    ), 1), INFINITY)
  );
  ctx->addSymbol("setsplit", 
    new ASTFunctionType(
      new ASTFunctionType(new ASTTypeVariable("_setsplit"), new ASTBaseType("bool"), INFINITY), 
      new ASTFunctionType(
        new ASTSetType(new ASTTypeVariable("_setsplit"), false), 
        new ASTTensorType(
          new ASTSetType(new ASTTypeVariable("_setsplit"), false),
          new ASTSetType(new ASTTypeVariable("_setsplit"), false)
    ), 1), INFINITY)
  );
  ctx->addSymbol("setmap", 
    new ASTFunctionType(
      new ASTFunctionType(new ASTTypeVariable("_setmap1"), new ASTTypeVariable("_setmap2"), INFINITY), 
      new ASTFunctionType(
        new ASTSetType(new ASTTypeVariable("_setmap1"), false), 
        new ASTSetType(new ASTTypeVariable("_setmap2"), false), 
    1), INFINITY)
  );
  ctx->addSymbol("bagmap", 
    new ASTFunctionType(
      new ASTFunctionType(new ASTTypeVariable("_bagmap1"), new ASTTypeVariable("_bagmap2"), INFINITY), 
      new ASTFunctionType(
        new ASTSetType(new ASTTypeVariable("_bagmap1"), true), 
        new ASTSetType(new ASTTypeVariable("_bagmap2"), true), 
    1), INFINITY)
  );
#endif
  ctx->addSymbol("cswp", 
    new ASTFunctionType(
      new ASTTensorType(new ASTBaseType("num"), new ASTBaseType("num")), 
      new ASTTensorType(new ASTBaseType("num"), new ASTBaseType("num")), 
    1)
  );
  ctx->addSymbol("setsum", 
    new ASTFunctionType(
      new ASTSetType(new ASTTypeVariable("clipped"), false),
      new ASTBaseType("num"),
    1)
  );
  ctx->addSymbol("bagsum", 
    new ASTFunctionType(
      new ASTSetType(new ASTTypeVariable("clipped"), true),
      new ASTBaseType("num"),
    1)
  );
  ctx->addSymbol("random", 
    new ASTFunctionType(
      new ASTBaseType("num"), 
      new ASTFunctionType(new ASTBaseType("num"), new ASTBaseType("num"), 
    INFINITY), INFINITY)
  );
  ctx->addSymbol("num2string", 
    new ASTFunctionType(
      new ASTBaseType("num"), 
      new ASTBaseType("string"), 
    1)
  );
        
  TCColons *colons = new TCColons();
  ASTType *typ = ast->getType(ctx, colons);

  if (ctx->getNumErrors() > 0) {
    fprintf(stderr, "Type check failed\n");
    return 2;
  } else if (ctx->getNumWarnings() > 0) {
    fprintf(stderr, "There were %d warnings(s) during the type check\n", ctx->getNumWarnings());
  }

  assert(typ);

  if (!datafileName && !outfileName) {
    printf("%s\n", render(typ));
    return 0;
  }

  if (verboseMode)
    printf("%s\n", render(typ));

  if (typ->getTypeCode() != TYP_LAMBDA)
    panic("Expression does not evaluate to a lambda type!");
    
  ASTFunctionType *ft = (ASTFunctionType*) typ;
  double sensitivity = ft->getSensitivity();
  if (!allowUnsampled && (ft->getResultType()->getTypeCode() != TYP_FUZZY))
    panic("Expression does not return a fuzzy type");
  
  if (datafileName) {
  
    /* Read data */
  
    ASTExpression *arg = readDataFile(datafileName, dataFormat);

    if (verboseMode) {  
      printf("\n=== Evaluation:\n");
      printf("Input: %s\n", render(arg));
    }
  
    /* Type check data */
  
    ASTType *dataType = arg->getType(new TCContext(), new TCColons());
    if (!dataType->equals(ft->getArgType(), ctx)) 
      panic("Data has type '%s', but we need type '%s'", render(dataType), render(ft->getArgType()));

    /* Evaluate */
  
    ast = ast->substitute("fuzz", new ASTBuiltinExpression("fuzz"));
    ast = ast->substitute("return", new ASTBuiltinExpression("return"));
    ast = ast->substitute("bagsize", new ASTBuiltinExpression("bagsize"));
    ast = ast->substitute("clip", new ASTBuiltinExpression("clip"));
    ast = ast->substitute("bagsplit", new ASTBuiltinExpression("bagsplit"));
    ast = ast->substitute("cswp", new ASTBuiltinExpression("cswp"));
    ast = ast->substitute("setsum", new ASTBuiltinExpression("setsum"));
    ast = ast->substitute("bagsum", new ASTBuiltinExpression("bagsum"));
    ast = ast->substitute("setmap", new ASTBuiltinExpression("setmap"));
    ast = ast->substitute("bagmap", new ASTBuiltinExpression("bagmap"));
    ast = ast->substitute("setsplit", new ASTBuiltinExpression("setsplit"));
    ast = ast->substitute("random", new ASTBuiltinExpression("random"));
    ast = ast->substitute("num2string", new ASTBuiltinExpression("num2string"));
  
    ASTExpression *result = (new ASTApplyExpression(ast, arg))->evaluate();
    assert(result);
  
    if (!allowUnsampled) {
      assert(result->getExpCode() == EXP_DISTRIBUTION);
      result = ((ASTDistribution*)result)->sample();
    }

    if (verboseMode)
      printf("Result: %s\n", render(result));
    else 
      printf("%s\n", render(result));
  }
  
  if (outfileName) {
    FILE *outfile = fopen(outfileName, "w+");
    if (!outfile)
      panic("Cannot open for writing: '%s'", outfileName);
      
    Printer *outPrinter = new Printer(outfile);
    BECamlVisitor *camlVisitor = new BECamlVisitor(outPrinter);
    outPrinter->print("#open \"item\";;\n let query = \n");
    ast->accept(camlVisitor);
    outPrinter->print(";;\n");
    fclose(outfile);
  }

  return 0;
}
