#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/wait.h>

#define panic(a...) do { fprintf(stderr, a); fprintf(stderr, "\n"); abort(); } while (0)

#define SRCFILENAME "__tmpsource.fuzz"

char currentSection[200];
bool sectionOpen = false;
int testsInSection;
int failedInSection;
int totalTests;
int totalFailed;
char nextCmd[200];
char nextPrg[2000];
char nextArg[2000];
bool debug = false;

void startSection(const char *title)
{
  strcpy(currentSection, title);
  testsInSection = 0;
  failedInSection = 0;
  sectionOpen = true;
  
  printf("%s ", currentSection);
  for (int i=strlen(currentSection)+1; i<40; i++)
    printf(".");
  printf(" ");
}

void finishSection()
{
  if (!failedInSection)
    printf("[ OK ]\n");
}

void subtestResult(const char *cmd, const char *prg, const char *arg, const char *output, const char *expected)
{
  if (strcmp(output, expected)) {
    if (!failedInSection)
      printf("[FAIL]\n");
    printf("Attempted: %s %s (%s)\n", cmd, prg, arg);
    printf("Result:    %s\n", output);
    printf("Expected:  %s\n", expected);
    failedInSection ++;
    totalFailed ++;
  }
  
  totalTests ++;
}

int main(int argc, char *argv[])
{
  if (argc != 2) 
    panic("Syntax: testfuzz <testfile>");
    
  FILE *infile = fopen(argv[1], "r");
  if (!infile)
    panic("Cannot open for reading: '%s'", argv[1]);

  strcpy(currentSection, "Main");
  testsInSection = 0;
  nextCmd[0] = 0;
  
  char linebuf[2000];
  while (fgets(linebuf, sizeof(linebuf), infile)) {
    strtok(linebuf, "#\r\n");
    if (linebuf[0] == '#')
      linebuf[0] = 0;
      
    char *head = strtok(linebuf, " ");
    char *rest = head ? strtok(NULL, "\r") : NULL;
    if (!rest)
      continue;
      
    while (rest[0] == ' ') {
      for (int p=0; rest[p]; p++)
        rest[p] = rest[p+1];
    }
    
    while (rest[0] && rest[strlen(rest)-1] == ' ')
      rest[strlen(rest)-1] = 0;
      
    if (!strcasecmp(head, "section")) {
      if (sectionOpen)
        finishSection();
        
      startSection(rest);
    } else if (!strcasecmp(head, "type") || !strcasecmp(head, "eval")) {
      if (nextCmd[0])
        panic("'type' and 'eval' must be followed by 'result'");
        
      strncpy(nextCmd, head, sizeof(nextCmd));
      strncpy(nextPrg, rest, sizeof(nextPrg));
    } else if (!strcasecmp(head, "arg")) {
      strncpy(nextArg, rest, sizeof(nextArg));
    } else if (!strcasecmp(head, "result")) {
      if (!sectionOpen) 
        startSection(currentSection);
   
      if (!nextCmd[0])
        panic("'result' must be preceded by a 'type' or 'eval'");
        
      FILE *outfile = fopen(SRCFILENAME, "w+");
      if (!outfile)
        panic("Cannot open for writing: '%s'", SRCFILENAME);
      
      fprintf(outfile, "%s\n", nextPrg);
      fclose(outfile);

      if (debug)
        fprintf(stderr, "---\nProgram is '%s'\n", nextPrg);
        
      char *margv[50];
      int p = 0;
      margv[p++] = (char*)"fuzz";
      margv[p++] = (char*)"--permissive";
      margv[p++] = (char*)"--nonoise";
      if (!strcasecmp(nextCmd, "eval")) {
        margv[p++] = (char*)"--format";
        if (strstr(nextArg, "|"))
          margv[p++] = (char*)"numlist";
        else
          margv[p++] = (char*)"num";
        margv[p++] = (char*)"--eval";
        margv[p++] = (char*)"-";
      }
      margv[p++] = (char*)SRCFILENAME;
        
      margv[p] = NULL;
        
      if (debug) {
        fprintf(stderr, "Running");
        for (int i=0; i<p; i++)
          fprintf(stderr, " %s", margv[i]);
        fprintf(stderr, "\n");
      }
      
      int downpipe[2];
      pipe(downpipe);
      int uppipe[2];
      pipe(uppipe);
        
      pid_t childPID = fork();
      if (childPID == 0) {
        close(STDIN_FILENO);
        if (!debug)
         close(STDERR_FILENO);
        dup2(downpipe[0], STDIN_FILENO);
        close(STDOUT_FILENO);
        dup2(uppipe[1], STDOUT_FILENO);
        execvp("./fuzz", margv);
        panic("Cannot fork 'fuzz' process (%s)", strerror(errno));
      } else {
        close(downpipe[0]);
        close(uppipe[1]);
      }
      
      if (!strcasecmp(nextCmd, "eval")) {
        if (debug)
          fprintf(stderr, "Sending input '%s'\n", nextArg);
          
        if (strstr(nextArg, "|")) {
          char *t = strtok(nextArg, "|");
          while (t) {
            write(downpipe[1], t, strlen(t));
            write(downpipe[1], "\r\n", 2);
            t = strtok(NULL, "|\r\n");
          }
        } else {
          write(downpipe[1], nextArg, strlen(nextArg));
        }
        
        write(downpipe[1], "\r\n", 2);
      }
        
      close(downpipe[1]);
      
      char inbuf[5000];
      int nread = read(uppipe[0], inbuf, sizeof(inbuf));
      if (nread < 0)
        panic("Cannot read from child program (%s)", strerror(errno));
        
      strtok(inbuf, "\r\n");
      
      if (debug) {
        fprintf(stderr, "Read output '%s'\n", inbuf);
        fprintf(stderr, "Expected output '%s'\n", rest);
      }
        
      int status;
      pid_t child2 = wait(&status);
      if (!strcasecmp(rest, "<err>")) {
        if (status == 0)
          sprintf(inbuf, "<Exits normally>");
        else if (status != 512)
          sprintf(inbuf, "<Exits with unexpected code %d>", status);
        else
          sprintf(inbuf, "<err>");
      } else {
        if (status == 512)
          sprintf(inbuf, "<err>");
        else if (status != 0)
          sprintf(inbuf, "<Exits with status code %d>", status);
      }

      subtestResult(nextCmd, nextPrg, nextArg, inbuf, rest);
        
      nextCmd[0] = 0;
      nextArg[0] = 0;
      nextPrg[0] = 0;
      unlink(SRCFILENAME);
      
    } else {
      panic("Unknown command: '%s'", head);
    }
  }
    
  finishSection();
  
  printf("\n%d test%s successful", totalTests-totalFailed, (totalTests-totalFailed)==1 ? "" : "s");
  if (totalFailed)
    printf(", %d test%s failed", totalFailed, (totalFailed==1) ? "" : "s");
  printf("\n");
    
  fclose(infile);
  
  return 0;
}
