#include <math.h> 

#include "global.h"
#include "dump.h"

char *prettyDouble(double d, char *buf)
{
  sprintf(buf, "%.5f", d);
  while ((strlen(buf)>0) && (buf[strlen(buf)-1] == '0'))
    buf[strlen(buf)-1] = 0;
  if ((strlen(buf)>0) && (buf[strlen(buf)-1] == '.'))
    buf[strlen(buf)-1] = 0;
  return buf;
}

ASTDumpVisitor::ASTDumpVisitor(Printer *printer, bool singleMode)
{
  this->printer = printer;
  this->singleMode = singleMode;
}

ASTDumpVisitor::~ASTDumpVisitor()
{
  delete printer;
}

void ASTDumpVisitor::visit(ASTConstNumber *peer)
{
  char buf[200];
  printer->print("%s", prettyDouble(peer->getValue(), buf));
}

void ASTDumpVisitor::visit(ASTConstBool *peer)
{
  printer->print("%s", peer->getValue() ? "true" : "false");
}

void ASTDumpVisitor::visit(ASTConstString *peer)
{
  printer->print("%s", peer->getValue());
}

void ASTDumpVisitor::visit(ASTLetBindingExpression *peer)
{
  printer->print("%s = ", peer->getNameToBind());
  peer->getRightHandExp()->accept(this);
  printer->println(";");
  if (singleMode)
    printer->print("...");
  else
    peer->getFollowExp()->accept(this);
}

void ASTDumpVisitor::visit(ASTPrintExpression *peer)
{
  printer->print("print ");
  peer->getRightHandExp()->accept(this);
  printer->println(";");
  if (singleMode)
    printer->print("...");
  else
    peer->getFollowExp()->accept(this);
}

void ASTDumpVisitor::visit(ASTTypeDefExpression *peer)
{
  printer->print("%s = ", peer->getNameToBind());
  peer->getTypeDef()->accept(this);
  printer->println(";");
  peer->getFollowExp()->accept(this);
}

void ASTDumpVisitor::visit(ASTLambdaExpression *peer)
{
  char buf[200];
  printer->print("fun ");
  if (peer->getOwnNameOrNull())
    printer->print("%s ", peer->getOwnNameOrNull());
  printer->print("(%s :", peer->getNameToBind());
  if (peer->getDeclaredSensitivity() == 1)
    printer->print("[]");
  else if (!isinf(peer->getDeclaredSensitivity()))
    printer->print("[%s]", prettyDouble(peer->getDeclaredSensitivity(), buf));
  printer->print(" ");
  peer->getBindingType()->accept(this);
  printer->print(") ");
  if (peer->getReturnTypeOrNull()) {
    printer->print(": ");
    peer->getReturnTypeOrNull()->accept(this);
    printer->print(" ");
  }
  printer->println("{");
  printer->indent(+1);
  if (!singleMode)
    peer->getInnerExp()->accept(this);
  else
    printer->print("...");
  printer->println();
  printer->indent(-1);
  printer->print("}");
}

void ASTDumpVisitor::visit(ASTBaseType *peer)
{
  printer->print("%s", peer->getName());
}

void ASTDumpVisitor::visit(ASTInfixOpExpression *peer)
{
  peer->getLeftExpression()->accept(this);
  printer->print(" %s ", peer->getOp());
  peer->getRightExpression()->accept(this);
}

void ASTDumpVisitor::visit(ASTVariable *peer)
{
  printer->print("%s", peer->getName());
}

void ASTDumpVisitor::visit(ASTUnitValue *peer)
{
  printer->print("()");
}

void ASTDumpVisitor::visit(ASTTypeVariable *peer)
{
  printer->print("%s", peer->getName());
  if (peer->getSequenceNo() != 0)
    printer->print("_%d", peer->getSequenceNo());
}

void ASTDumpVisitor::visit(ASTNamedType *peer)
{
  printer->print("%s", peer->getName());
}

void ASTDumpVisitor::visit(ASTUnionType *peer)
{
  peer->getOption(0)->accept(this);
  printer->print(" + ");
  peer->getOption(1)->accept(this);
}

#define acceptWithParens(x) do { if ((x)->needsParentheses()) printer->print("("); (x)->accept(this); if ((x)->needsParentheses()) printer->print(")"); } while (0)

void ASTDumpVisitor::visit(ASTTensorType *peer)
{
  acceptWithParens(peer->getLeft());
  printer->print(" * ");
  acceptWithParens(peer->getRight());
}

void ASTDumpVisitor::visit(ASTAmpersandType *peer)
{
  peer->getLeft()->accept(this);
  printer->print(" & ");
  peer->getRight()->accept(this);
}

void ASTDumpVisitor::visit(ASTFuzzyType *peer)
{
  printer->print("fuzzy ");
  peer->getInnerType()->accept(this);
}

void ASTDumpVisitor::visit(ASTMuType *peer)
{
  printer->print("mu %s => ", peer->getName());
  peer->getInnerType()->accept(this);
}

void ASTDumpVisitor::visit(ASTCaseExpression *peer)
{
  printer->print("case ");
  peer->getExpr()->accept(this);
  printer->println(" {");
  printer->indent(+1);
  printer->println("inl(%s) =>", peer->getId(0));
  printer->indent(+1);
  peer->getBranch(0)->accept(this);
  printer->indent(-1);
  printer->println();
  printer->println("inr(%s) =>", peer->getId(1));
  printer->indent(+1);
  peer->getBranch(1)->accept(this);
  printer->indent(-1);
  printer->println();
  printer->indent(-1);
  printer->print("}");
}

void ASTDumpVisitor::visit(ASTIfThenElseExpression *peer)
{
  printer->print("if ");
  peer->getExpr()->accept(this);
  printer->indent(+1);
  printer->println("then ");
  printer->indent(+1);
  peer->getThenBranch()->accept(this);
  printer->indent(-1);
  printer->println();
  printer->println("else ");
  printer->indent(+1);
  peer->getElseBranch()->accept(this);
  printer->indent(-1);
  printer->println();
}

void ASTDumpVisitor::visit(ASTUnfoldExpression *peer)
{
  printer->print("unfold ");
  peer->getInnerExp()->accept(this);
}

void ASTDumpVisitor::visit(ASTBuiltinExpression *peer)
{
  printer->print("%s", peer->getName());
}

extern bool enablePrettyPrinting;

bool isList(ASTType *elementType, ASTExpression *value)
{
  if (value->getExpCode() != EXP_FOLD)
    return false;

  ASTFoldExpression *foldExp = (ASTFoldExpression*) value;
  if (foldExp->getInnerExp()->getExpCode() != EXP_INJ)
    return false;

  ASTInjectExpression *injExp = (ASTInjectExpression*) foldExp->getInnerExp();
  if ((injExp->getBranch() == 0) && (injExp->getInnerExp()->getExpCode() == EXP_UNIT))
    return true;

  if ((injExp->getBranch() == 1) && (injExp->getInnerExp()->getExpCode() == EXP_TENSOR)) {
    ASTTensorTuple *tup = (ASTTensorTuple*) injExp->getInnerExp();
    return isList(elementType, tup->getRightExpression());
  }

  return false;
}

void ASTDumpVisitor::visit(ASTFoldExpression *peer)
{
  ASTType *listType = NULL;
  bool isTensor = false;
  if (peer->getTypeArg()->getTypeCode() == TYP_MU) {
    ASTMuType *muType = (ASTMuType*) peer->getTypeArg();
    if (muType->getInnerType()->getTypeCode() == TYP_UNION) {
      ASTUnionType *uni = (ASTUnionType*) muType->getInnerType();
      int unitBranch = 0;
      if (uni->getOption(unitBranch)->getTypeCode() != TYP_UNIT)
        unitBranch = 1;
      if (uni->getOption(unitBranch)->getTypeCode() == TYP_UNIT) {
        if (uni->getOption(1-unitBranch)->getTypeCode() == TYP_TENSOR) {
          ASTTensorType *tens = (ASTTensorType*) uni->getOption(1-unitBranch);
          if (tens->getRight()->getTypeCode() == TYP_VAR) {
            listType = tens->getLeft();
            isTensor = true;
          }
        } else if (uni->getOption(1-unitBranch)->getTypeCode() == TYP_AMPERSAND) {
          ASTAmpersandType *amp = (ASTAmpersandType*) uni->getOption(1-unitBranch);
          if (amp->getRight()->getTypeCode() == TYP_VAR) {
            listType = amp->getLeft();
            isTensor = false;
          }
        } 
      }
    }
  }

  if (listType && enablePrettyPrinting && isList(listType, peer)) {
    printer->print("(");
    ASTExpression *iter = peer;
    while (true) {
      ASTExpression *olditer = iter;
      assert(iter->getExpCode() == EXP_FOLD);
      iter = ((ASTFoldExpression*)iter)->getInnerExp();
      assert(iter->getExpCode() == EXP_INJ);
      ASTInjectExpression *injExp = (ASTInjectExpression*)iter;
      if (injExp->getBranch() == 0)
        break;
      if (olditer != peer)
        printer->print(",");
      else 
        printer->print(" ");
      assert(injExp->getInnerExp()->getExpCode() == EXP_TENSOR);
      ASTTensorTuple *tens = (ASTTensorTuple*) injExp->getInnerExp();
      tens->getLeftExpression()->accept(this);
      iter = tens->getRightExpression();
    }
    printer->print(" )");
  } else {
    printer->print("fold[");
    peer->getTypeArg()->accept(this);
    printer->print("] ");
    peer->getInnerExp()->accept(this);
  }
}

void ASTDumpVisitor::visit(ASTInjectExpression *peer)
{
  printer->print("in%c[", peer->getBranch() ? 'r' : 'l');
  peer->getTypeArg()->accept(this);
  printer->print("] ");
  peer->getInnerExp()->accept(this);
}

void ASTDumpVisitor::visit(ASTProjectExpression *peer)
{
  printer->print("%s ", peer->getBranch() ? "snd" : "fst");
  peer->getInnerExp()->accept(this);
}

void ASTDumpVisitor::visit(ASTTensorTuple *peer)
{
  printer->print("(");
  peer->getLeftExpression()->accept(this);
  printer->print(",");
  peer->getRightExpression()->accept(this);
  printer->print(")");
}

void ASTDumpVisitor::visit(ASTTensorDestructorExpression *peer)
{
  printer->print("let (%s,%s) = ", peer->getName(0), peer->getName(1));
  peer->getProducingExp()->accept(this);
  printer->println(";");
  peer->getConsumingExp()->accept(this);
}

void ASTDumpVisitor::visit(ASTAmpersandTuple *peer)
{
  printer->print("(|");
  peer->getLeftExpression()->accept(this);
  printer->print(",");
  peer->getRightExpression()->accept(this);
  printer->print("|)");
}

void ASTDumpVisitor::visit(ASTApplyExpression *peer)
{
    printer->print("(");
  peer->getFun()->accept(this);
  printer->print(" ");
//  if (peer->getArg()->needsParentheses()) 
  peer->getArg()->accept(this);
//  if (peer->getArg()->needsParentheses()) 
    printer->print(")");
}

void ASTDumpVisitor::visit(ASTUnitType *peer)
{
  printer->print("()");
}

void ASTDumpVisitor::visit(ASTFunctionType *peer)
{
  char buf[200];
  if (peer->getArgType()->needsParentheses())
    printer->print("(");
  peer->getArgType()->accept(this);
  if (peer->getArgType()->needsParentheses())
    printer->print(")");
    
  if (!isinf(peer->getSensitivity())) {
    if (peer->getSensitivity() == 1)
      printer->print(" -o ");
    else
      printer->print(" -o[%s] ", prettyDouble(peer->getSensitivity(), buf));
  } else {
    printer->print(" -> ");
  }

  if (peer->getResultType()->needsParentheses())
    printer->print("(");
  peer->getResultType()->accept(this);
  if (peer->getResultType()->needsParentheses())
    printer->print(")");
}

void ASTDumpVisitor::visit(ASTSetType *peer)
{
  acceptWithParens(peer->getMemberType());
  printer->print(peer->isMultiset() ? " bag" : " set");
}

void ASTDumpVisitor::visit(ASTSetExpression *peer)
{
  printer->print("{ ");
  struct ASTSetExpression::elementStruct *iter = peer->elements;
  while (iter) {
    iter->value->accept(this);
    if (iter->next)
      printer->print(", ");
    iter = iter->next;
  }
  
  printer->print(" }");
}

void ASTDumpVisitor::visit(ASTSampleExpression *peer)
{
  printer->print("sample %s = ", peer->getNameToBind());
  peer->getRightHandExp()->accept(this);
  printer->println(";");
  peer->getFollowExp()->accept(this);
}

void ASTDumpVisitor::visit(ASTSingleValueDistribution *peer)
{
  char buf[200];
  printer->print("singleValueDist(");
  peer->getCenter()->accept(this);
  printer->print(")");
}

void ASTDumpVisitor::visit(ASTLaplaceDistribution *peer)
{
  char buf[200];
  printer->print("laplaceDist(%s)", prettyDouble(peer->getCenter(), buf));
}
