#include <stdarg.h>
#include <stdio.h>
#include <assert.h>

#include "global.h"
#include "ast.h"

//#define dprintf(a...) printf(a)
#define dprintf(a...) do { } while (0)

static int lastTypeVarSequenceNo = 0;

TCContext::TCContext()
{
  this->symbols = NULL;
  this->numErrors = 0;
  this->numWarnings = 0;
  this->nextHandle = 1;
  this->boundTypeVars = NULL;
  this->nextScopeHandle = 1;
}

void ASTExpression::showMessage(const char *buf)
{
  if (context)
    context->showMessage(buf);
  else
    printf("%s\n", buf);
}

void TCContext::addSymbol(const char *name, ASTType *type)
{
  TCSymbol *sym = new TCSymbol(name, type);
  sym->enqueueBefore(symbols);
  symbols = sym;
}

int TCContext::storeType(ASTType *type)
{
  int thisHandle = nextHandle ++;
  TCSymbol *sym = new TCSymbol(NULL, type, thisHandle);
  sym->enqueueBefore(symbols);
  symbols = sym;
  return thisHandle;
}

ASTType *TCContext::retrieveType(int handle)
{
  if (!symbols)
    return NULL;
    
  TCSymbol *sym = symbols->lookupByHandle(handle, symbols);
  if (symbols == sym) {
    if (sym->getNext() == sym) {
      symbols = NULL;
    } else {
      symbols = symbols->getNext();
    }
  }
  
  sym->dequeue();
  return sym->getType();
}

void TCContext::removeSymbol(const char *name)
{
  TCSymbol *sym = getSymbol(name);
  assert(sym);
  if (symbols == sym) {
    if (sym->getNext() == sym) {
      symbols = NULL;
    } else {
      symbols = symbols->getNext();
    }
  }
  
  sym->dequeue();
}

char *TCContext::render()
{
  const int maxlen = 20000;
  char *buf = (char*) malloc(maxlen);
  
  if (!symbols) {
    strcpy(buf, "{ <empty> }");
    return buf;
  }
  
  strcpy(buf, "{ ");
  symbols->render(buf, symbols, maxlen);
  strcat(buf, " }");
  
  return buf;
}

void TCContext::addError(ASTExpression *expr, const char *fmt, ...)
{
  char buf[2000];
  va_list vl;
  va_start(vl, fmt);
  
  int printed = vsnprintf(buf, sizeof(buf), fmt, vl);
  expr->showMessage(buf);

  va_end(vl);
  numErrors ++;
}

void TCContext::addWarning(ASTExpression *expr, const char *fmt, ...)
{
  char buf[2000];
  va_list vl;
  va_start(vl, fmt);
  
  int printed = vsnprintf(buf, sizeof(buf), fmt, vl);
  expr->showMessage(buf);

  va_end(vl);
  numWarnings ++;
}

ASTType *replaceTypeVar(ASTType *where, ASTTypeVariable *what, ASTType *replacement);

void TCContext::replaceTypeVarEverywhere(ASTTypeVariable *var, ASTType *replacement)
{
  if (symbols) {
    TCSymbol *iter = symbols;
    do { 
      if (iter->getType()) 
        iter->setType(replaceTypeVar(iter->getType(), var, replacement));

      iter = iter->getNext();
    } while (iter != symbols);
  }
}

class BagOfMuTypes {
  struct entry {
    const char *muVariable;
    struct entry *next;
  } *entries;
  
public:
  BagOfMuTypes()
    { this->entries = NULL; };
  void add(const char *var) {
    struct entry *item = (struct entry*) malloc(sizeof(struct entry));
    item->muVariable = var;
    item->next = entries;
    entries = item;
  }
  void remove(const char *var) {
    if (!entries)
      return;
    if (!strcmp(entries->muVariable, var)) {
      entries = entries->next;
      return;
    }
    struct entry *iter;
    while (iter->next) {
      if (!strcmp(iter->next->muVariable, var)) {
        iter->next = iter->next->next;
        return;
      }
      iter = iter->next;
    }
  }
  bool contains(const char *var) {
    struct entry *iter = entries;
    while (iter) {
      if (!strcmp(iter->muVariable, var))
        return true;
      iter = iter->next;
    }
    return false;
  }
};

int TCContext::bindAnyNewTypeVarsIn(ASTType *theType)
{
  class BindTypeVariablesVisitor : public ASTDFSVisitor {
    BagOfMuTypes *muTypes;
    TCContext *ctx;
    int numBound;
    int scopeHandle;

  public:
    BindTypeVariablesVisitor(TCContext *ctx, int scopeHandle) : ASTDFSVisitor()
      { this->ctx = ctx; this->numBound = 0; this->scopeHandle = scopeHandle; this->muTypes = new BagOfMuTypes(); };
    virtual void visit(ASTTypeVariable *peer) 
      { if (!ctx->isTypeVariableBound(peer) && !(muTypes->contains(peer->getName()) && (peer->getSequenceNo()==0))) { ctx->bindTypeVariableToScope(peer, scopeHandle); numBound ++; } };
    virtual void visit(ASTMuType *peer) 
      { muTypes->add(peer->getName()); ASTDFSVisitor::visit(peer); muTypes->remove(peer->getName()); };
    int getNumBound()
      { return numBound; };
  };

  int thisScopeHandle = nextScopeHandle ++;
  BindTypeVariablesVisitor *btvv = new BindTypeVariablesVisitor(this, thisScopeHandle);
  theType->accept(btvv);
  return thisScopeHandle;
}

bool TCContext::isTypeVariableBound(ASTTypeVariable *typeVar)
{
  struct boundTypeVarEntry *iter = boundTypeVars;
  while (iter) {
    if (iter->typeVar->equals(typeVar))
      return true;
    iter = iter->next;
  }
  return false;
}

void TCContext::bindTypeVariableToScope(ASTTypeVariable *typeVar, int handle)
{
  assert(!isTypeVariableBound(typeVar));
  dprintf("Binding type variable %s to scope #%d\n", ::render(typeVar), handle);
  struct boundTypeVarEntry *item = (struct boundTypeVarEntry*) malloc(sizeof(struct boundTypeVarEntry));
  item->typeVar = typeVar->clone();
  item->scopeHandle = handle;
  item->next = boundTypeVars;
  boundTypeVars = item;
}

void TCContext::unbindTypeVarsIn(int scopeHandle)
{
  if (!boundTypeVars)
    return;
    
  if (boundTypeVars->scopeHandle == scopeHandle) {
    struct boundTypeVarEntry *item = boundTypeVars;
    boundTypeVars = item->next;
    
    ASTTypeVariable *newTypeVar = item->typeVar->clone();
    newTypeVar->setSequenceNo(++ lastTypeVarSequenceNo);
    replaceTypeVarEverywhere(item->typeVar, newTypeVar);
    dprintf("Unbinding [%s]->[%s] in scope #%d\n", ::render(item->typeVar), ::render(newTypeVar), scopeHandle);
    
    unbindTypeVarsIn(scopeHandle);
    return;
  }
    
  struct boundTypeVarEntry *iter = boundTypeVars;
  while (iter->next != NULL) {
    if (iter->next->scopeHandle == scopeHandle) {
      struct boundTypeVarEntry *item = iter->next;
      iter->next = item->next;

      ASTTypeVariable *newTypeVar = item->typeVar->clone();
      newTypeVar->setSequenceNo(++ lastTypeVarSequenceNo);
      replaceTypeVarEverywhere(item->typeVar, newTypeVar);
      dprintf("Unbinding [%s]->[%s] in scope #%d\n", ::render(item->typeVar), ::render(newTypeVar), scopeHandle);

      unbindTypeVarsIn(scopeHandle);
      return;
    }
    
    iter = iter->next;
  }
}

ASTType *TCContext::getSymbolType(const char *name)
{
  class FreshTypeVariablesVisitor : public ASTDFSVisitor {
    struct replacementEntryStruct {
      ASTTypeVariable *var;
      int newSeq;
      struct replacementEntryStruct *next;
    } *replacements;
    TCContext *daddy;
    BagOfMuTypes *muTypes;

  public:
    FreshTypeVariablesVisitor(TCContext *daddy) : ASTDFSVisitor()
      { replacements = NULL; this->daddy = daddy; muTypes = new BagOfMuTypes(); };
    int getNewSeqFor(ASTTypeVariable *peer) {
      struct replacementEntryStruct *iter = replacements;
      while (iter) {
        if (iter->var->equals(peer))
          return iter->newSeq;
        iter = iter->next;
      };
      return -1;
    }
    virtual void visit(ASTMuType *peer) 
      { muTypes->add(peer->getName()); ASTDFSVisitor::visit(peer); muTypes->remove(peer->getName()); };
    virtual void visit(ASTTypeVariable *peer) {
      if (!daddy->isTypeVariableBound(peer) && !(muTypes->contains(peer->getName()) && (peer->getSequenceNo()==0))) {
        int newSeq = getNewSeqFor(peer);
        if (newSeq < 0) {
          newSeq = ++lastTypeVarSequenceNo;
          struct replacementEntryStruct *item = (struct replacementEntryStruct*) malloc(sizeof(struct replacementEntryStruct));
          item->var = peer->clone();
          item->newSeq = newSeq;
          item->next = replacements;
          replacements = item;
        }
        peer->setSequenceNo(newSeq);
      }
    }
  };

  TCSymbol *sym = getSymbol(name);
  if (!sym)
    return NULL;
  
  ASTType *result = sym->getType()->clone();  
  result->accept(new FreshTypeVariablesVisitor(this));
  
  return result;
}
