#include "ast.h"

#define differenceFound(mine, peer) do { assert(equalSoFar); equalSoFar = false; firstDifference.localSubtree = (mine); firstDifference.peerSubtree = (peer); return; } while (0)

class TCEqualityVisitor : public ASTVisitor
{
protected:
  TCContext *ctx;
  TCVariableMapping *mapping;
  ASTType *peerNode;
  bool equalSoFar; 
  struct {
    ASTType *localSubtree;
    ASTType *peerSubtree;
  } firstDifference;
  
public:
  TCEqualityVisitor(TCContext *ctx, bool ignoreRenaming, ASTType *peerNode);
  virtual void visit(ASTBaseType *myNode);
  virtual void visit(ASTTypeVariable *myNode);
  virtual void visit(ASTUnionType *myNode);
  virtual void visit(ASTTensorType *myNode);
  virtual void visit(ASTAmpersandType *myNode);
  virtual void visit(ASTFuzzyType *myNode);
  virtual void visit(ASTUnitType *myNode);
  virtual void visit(ASTFunctionType *myNode);
  virtual void visit(ASTMuType *myNode);
  virtual void visit(ASTNamedType *myNode);
  virtual void visit(ASTSetType *myNode);
  bool isEqual()
    { return equalSoFar; };
  ASTType *getFirstDifferenceLocal() 
    { return firstDifference.localSubtree; };
  ASTType *getFirstDifferencePeer()
    { return firstDifference.peerSubtree; };
};

TCEqualityVisitor::TCEqualityVisitor(TCContext *ctx, bool ignoreRenaming, ASTType *peerNode)
{
  this->ctx = ctx;
  this->mapping = ignoreRenaming ? new TCVariableMapping() : NULL;
  this->peerNode = peerNode;
  this->equalSoFar = true;
  this->firstDifference.localSubtree = NULL;
  this->firstDifference.peerSubtree = NULL;
}

void TCEqualityVisitor::visit(ASTBaseType *myNode)
{
  ASTType *peerResolved = peerNode->resolve(ctx); 
  if (peerResolved->getTypeCode() != TYP_BASE) 
    differenceFound(myNode, peerResolved);
  
  ASTBaseType *bt = (ASTBaseType*) peerResolved;
  if (strcmp(myNode->getName(), bt->getName()))
    differenceFound(myNode, peerResolved);
}

void TCEqualityVisitor::visit(ASTTypeVariable *myNode)
{
  ASTType *peerResolved = peerNode->resolve(ctx);
  if (peerResolved->getTypeCode() != TYP_VAR)
    differenceFound(myNode, peerResolved);
  
  ASTTypeVariable *var = (ASTTypeVariable*) peerResolved;
  const char *hisName = (mapping && mapping->mappingExists(var->getName())) ? mapping->lookupMapping(var->getName()) : var->getName();
  if (strcmp(myNode->getName(), hisName))
    differenceFound(myNode, peerResolved);
}

void TCEqualityVisitor::visit(ASTNamedType *myNode)
{
  assert(ctx->containsSymbol(myNode->getName()));

  /* Recursion */

  peerNode = peerNode->resolve(ctx);
  myNode->resolve(ctx)->accept(this);
}

void TCEqualityVisitor::visit(ASTUnionType *myNode)
{
  ASTType *peerResolved = peerNode->resolve(ctx);
  if (peerResolved->getTypeCode() != TYP_UNION) 
    differenceFound(myNode, peerResolved);

  /* Recursion */

  ASTUnionType *uni = (ASTUnionType*) peerResolved;
  peerNode = uni->getOption(0);
  myNode->getOption(0)->accept(this);
  
  if (!equalSoFar)
    return;
    
  peerNode = uni->getOption(1);
  myNode->getOption(1)->accept(this);
}

void TCEqualityVisitor::visit(ASTSetType *myNode)
{
  ASTType *peerResolved = peerNode->resolve(ctx);
  if (peerResolved->getTypeCode() != TYP_SET) 
    differenceFound(myNode, peerResolved);
  if (myNode->isMultiset() != ((ASTSetType*)peerResolved)->isMultiset())
    differenceFound(myNode, peerResolved);
  
  /* Recursion */

  ASTSetType *rset = (ASTSetType*) peerResolved;
  peerNode = rset->getMemberType();
  myNode->getMemberType()->accept(this);
}

void TCEqualityVisitor::visit(ASTTensorType *myNode)
{
  ASTType *peerResolved = peerNode->resolve(ctx);
  if (peerResolved->getTypeCode() != TYP_TENSOR) 
    differenceFound(myNode, peerResolved);

  /* Recursion */

  ASTTensorType *ten = (ASTTensorType*) peerResolved;
  peerNode = ten->getLeft();
  myNode->getLeft()->accept(this);
  
  if (!equalSoFar)
    return;
  
  peerNode = ten->getRight();
  myNode->getRight()->accept(this);
}

void TCEqualityVisitor::visit(ASTAmpersandType *myNode)
{
  ASTType *peerResolved = peerNode->resolve(ctx);
  if (peerResolved->getTypeCode() != TYP_AMPERSAND) 
    differenceFound(myNode, peerResolved);

  /* Recursion */

  ASTAmpersandType *amp = (ASTAmpersandType*) peerResolved;
  peerNode = amp->getLeft();
  myNode->getLeft()->accept(this);
  
  if (!equalSoFar) 
    return;

  peerNode = amp->getRight();
  myNode->getRight()->accept(this);
}

void TCEqualityVisitor::visit(ASTFuzzyType *myNode)
{
  ASTType *peerResolved = peerNode->resolve(ctx);
  if (peerResolved->getTypeCode() != TYP_FUZZY) 
    differenceFound(myNode, peerResolved);
  
  /* Recursion */

  ASTFuzzyType *ft = (ASTFuzzyType*) peerResolved;
  peerNode = ft->getInnerType();
  myNode->getInnerType()->accept(this);
}

void TCEqualityVisitor::visit(ASTUnitType *myNode)
{
  ASTType *peerResolved = peerNode->resolve(ctx); 
  if (peerResolved->getTypeCode() != TYP_UNIT) 
    differenceFound(myNode, peerResolved);
}

void TCEqualityVisitor::visit(ASTFunctionType *myNode)
{
  ASTType *peerResolved = peerNode->resolve(ctx);
  if (peerResolved->getTypeCode() != TYP_LAMBDA) 
    differenceFound(myNode, peerResolved);

  ASTFunctionType *ft = (ASTFunctionType*) peerResolved;
  if (myNode->getSensitivity() != ft->getSensitivity())
    differenceFound(myNode, peerResolved);
  
  /* Recursion */

  peerNode = ft->getArgType();
  myNode->getArgType()->accept(this);
  if (!equalSoFar)
    return;
    
  peerNode = ft->getResultType();
  myNode->getResultType()->accept(this);
}

void TCEqualityVisitor::visit(ASTMuType *myNode)
{
  ASTType *peerResolved = peerNode->resolve(ctx);
  if (peerResolved->getTypeCode() != TYP_MU) 
    differenceFound(myNode, peerResolved);
  
  ASTMuType *pmu = (ASTMuType*) peerResolved;
  if (mapping)
    mapping->addMapping(pmu->getName(), myNode->getName());
  
  peerNode = pmu->getInnerType();
  myNode->getInnerType()->accept(this);

  if (mapping)
    mapping->removeMapping(pmu->getName());
}

bool ASTType::equals(ASTType *peer, TCContext *ctx, ASTType **diffLocal, ASTType **diffPeer)
{ 
  TCVariableMapping *mapping = new TCVariableMapping();
  TCEqualityVisitor *visitor = new TCEqualityVisitor(ctx, true, peer);
  this->accept(visitor);
  
  if (!visitor->isEqual()) {
    if (diffLocal)
      *diffLocal = visitor->getFirstDifferenceLocal();
    if (diffPeer)
      *diffPeer = visitor->getFirstDifferencePeer();
  }
  
  return visitor->isEqual();
}
