#include <assert.h>

#include "global.h"
#include "ast.h"
#include "printer.h"

#define BEGIN_TYPE(exp, ctx, colons) do { \
  extern bool debugMode; \
  if (debugMode) { \
    if (!printer) \
      printer = new Printer(stdout); \
    printer->println("BeginType: [%s]", render(exp, true)); \
    printer->println("Ctxt:      %s", (ctx)->render()); \
    printer->println("Colons   : %s", (colons)->render()); \
  } \
} while (0)

#define END_TYPE(exp, typ, ctx, colons) do { \
  extern bool verboseMode, debugMode; \
  if (verboseMode) { \
    if (!printer) \
      printer = new Printer(stdout); \
    printer->println("Typing:  [%s]", render(exp, true)); \
    if ((exp)->getContext()) { \
      char buf[2000]; \
      strcpy(buf, (exp)->getContext()->line); \
      strtok(buf, "\r\n"); \
      char *p=buf; \
      while (*p == ' ') p++; \
      printer->println("At pos:  \"%s\" (%s:%d:%d)", p, (exp)->getContext()->file, (exp)->getContext()->lineNo, (exp)->getContext()->pos); \
    } else {\
      printer->println("At pos:  <unknown>"); \
    } \
    printer->println("Result:  [%s]", render(typ)); \
    if (debugMode) { \
      printer->println("Context: %s", (ctx)->render()); \
    } \
    printer->println("Colons : %s", (colons)->render()); \
    printer->println(); \
  } \
} while (0)

static Printer *printer = NULL;

ASTType *ASTConstNumber::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);
  
  ASTType *result = new ASTBaseType("num");
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTConstBool::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);
  
  ASTType *result = new ASTBaseType("bool");
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTConstString::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);
  
  ASTType *result = new ASTBaseType("string");
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTUnitValue::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);
  
  ASTType *result = new ASTUnitType();
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTLambdaExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  /* Add the variable we're binding, and possibly the name of the function, to the context */

  ctx->addSymbol(nameToBind, bindingType->clone());
  if (ownNameOrNull) {
    assert(returnTypeOrNull != NULL);
    ctx->addSymbol(ownNameOrNull, new ASTFunctionType(bindingType->clone(), returnTypeOrNull->clone(), declaredSensitivity));
  }
  
  int scopeHandle = ctx->bindAnyNewTypeVarsIn(bindingType);
  int bindingTypeHandle = ctx->storeType(bindingType->clone());
  int returnTypeHandleOrMinus1 = returnTypeOrNull ? ctx->storeType(returnTypeOrNull->clone()) : -1;
  
  /* Infer the inner type */
  
  TCColons *innerColons = new TCColons();
  ASTType *resultType = innerExp->getType(ctx, innerColons);
  if (!resultType) 
    return NULL;

  /* Extract the results and cleanup */
  
  double inferredSensitivity = innerColons->containsSymbol(nameToBind) ? innerColons->getSymbol(nameToBind) : 0;
  
  ctx->removeSymbol(nameToBind);
  if (ownNameOrNull) 
    ctx->removeSymbol(ownNameOrNull);
    
  /* Propagate the colons */
  
  innerColons->removeSymbol(nameToBind);
  colons->add(innerColons);
  
  /* Check whether the inferred sensitivity matches the declared sensitivity */
  
  if (inferredSensitivity > declaredSensitivity) {
    ctx->addError(this, "Lambda expression was declared with sensitivity [%.2f], but its sensitivity was inferred as %.2f", declaredSensitivity, inferredSensitivity);
    return NULL;
  }
  
  if ((enabledWarnings & WARN_TIGHT_SENSITIVITIES) && !SAME_SENSITIVITY(inferredSensitivity, declaredSensitivity)) {
    if (isinf(declaredSensitivity))
      ctx->addWarning(this, "Lambda expression was declared without [], but its sensitivity was inferred as %.2f", inferredSensitivity);
    else
      ctx->addWarning(this, "Lambda expression was declared with sensitivity [%.2f], but its sensitivity was inferred as %.2f", declaredSensitivity, inferredSensitivity);
  }

  /* If a return type was declared, check whether the inferred type matches */

  if (returnTypeHandleOrMinus1 >= 0) {
    ASTType *returnTypeBeforeUnification = ctx->retrieveType(returnTypeHandleOrMinus1);
    if (!ASTType::unify(&resultType, &returnTypeBeforeUnification, ctx)) {
      ctx->addError(this, "Lambda expression was supposed to return '%s', but actually returns '%s'", render(returnTypeBeforeUnification), render(resultType));
      return NULL;
    }
  }
  
  /* Construct the return type */
  
  int resultTypeHandle = ctx->storeType(resultType);
  ctx->unbindTypeVarsIn(scopeHandle);
  
  ASTType *result = new ASTFunctionType(ctx->retrieveType(bindingTypeHandle), ctx->retrieveType(resultTypeHandle), declaredSensitivity);

  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTVariable::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  /* Update the colon */

  if (!colons->containsSymbol(this->name))
    colons->addSymbol(this->name, 1);
  else
    colons->setSymbol(this->name, colons->getSymbol(this->name) + 1);
    
  /* Get the type, and replace all type variables with fresh ones */

  ASTType *result = ctx->getSymbolType(this->name);
  if (!result) {
    ctx->addError(this, "Cannot determine type of variable '%s'", this->name);
    return NULL;
  }
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTInfixOpExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  /* Typecheck both branches */

  TCColons *colonsLeft = new TCColons();
  int typLeftHandle = ctx->storeType(left->getType(ctx, colonsLeft));
  
  TCColons *colonsRight = new TCColons();
  ASTType *typRight = right->getType(ctx, colonsRight);
  ASTType *typLeft = ctx->retrieveType(typLeftHandle);
  
  if (!typLeft || !typRight) 
    return NULL;
  
  ASTType *result = NULL;
  
  if (!strcmp(op, "+") || !strcmp(op, "-") || !strcmp(op, "*") || !strcmp(op, "/")) {

    /* Numeric operators: num -> num -> num */
  
    ASTType *typLeftDesired = new ASTBaseType("num");
    ASTType *typRightDesired = new ASTBaseType("num");
    if (!ASTType::unify(&typLeft, &typLeftDesired, ctx)) {
      if (strcmp(op, "+")) {
        ctx->addError(this, "The left subexpression of '%s' cannot be unified with 'num' (%s)", op, render(typLeft));
        return NULL;
      } 
      
      typLeftDesired = new ASTBaseType("string");
      typRightDesired = new ASTBaseType("string");
      if (!ASTType::unify(&typLeft, &typLeftDesired, ctx)) {
        ctx->addError(this, "The left subexpression of '%s' cannot be unified with either 'num' or 'string' (%s)", op, render(typLeft));
        return NULL;
      } 
    }

    if (!ASTType::unify(&typRight, &typRightDesired, ctx)) {
      ctx->addError(this, "The left and right subexpressions of '%s' cannot be unified (%s vs %s)", op, render(typLeft), render(typRight));
      return NULL;
    }

    if (!strcmp(op, "+") || !strcmp(op, "-")) {
      colons->add(colonsLeft);
      colons->add(colonsRight);
    } else if (!strcmp(op, "*") || !strcmp(op, "/")) {
      if (left->getExpCode() == EXP_CONSTNUM) {
        colonsRight->multiplyBy(!strcmp(op, "*") ? (((ASTConstNumber*)left)->getValue()) : (1/((ASTConstNumber*)left)->getValue()));
        colons->add(colonsRight);
      } else if (right->getExpCode() == EXP_CONSTNUM) {
        colonsLeft->multiplyBy(!strcmp(op, "*") ? (((ASTConstNumber*)right)->getValue()) : (1/((ASTConstNumber*)right)->getValue()));
        colons->add(colonsLeft);
      } else {
        colonsRight->multiplyBy(INFINITY);
        colonsLeft->multiplyBy(INFINITY);
        colons->add(colonsRight);
        colons->add(colonsLeft);
      }
    } else {
      panic("Type inference for operator '%s' not implemented", op);
    }
    
    result = typLeft;
  } else if (!strcmp(op, "==") || !strcmp(op, "<") || !strcmp(op, ">") || !strcmp(op, "<=") || !strcmp(op, ">=")) {
  
    /* Comparison operators: A -> A -> bool */
    
    if (!ASTType::unify(&typLeft, &typRight, ctx))  {
      ctx->addError(this, "The left and right subexpressions of '%s' need to have the same type, but the left type is '%s' and the right type is '%s'", op, render(typLeft), render(typRight));
      return NULL;
    }
    
    colons->add(colonsLeft);
    colons->add(colonsRight);
    result = new ASTBaseType("bool");
    
  } else if (!strcmp(op, "&&") || !strcmp(op, "||")) {
  
    /* Logic operators: bool -> bool -> bool */

    ASTType *typLeftDesired = new ASTBaseType("bool");
    if (!ASTType::unify(&typLeft, &typLeftDesired, ctx)) {
      ctx->addError(this, "The left subexpressions of '%s' cannot be unified with 'bool' (%s)", op, render(typLeft));
      return NULL;
    }

    ASTType *typRightDesired = new ASTBaseType("bool");
    if (!ASTType::unify(&typRight, &typRightDesired, ctx)) {
      ctx->addError(this, "The left subexpressions of '%s' cannot be unified with 'bool' (%s)", op, render(typRight));
      return NULL;
    }
    
    colons->add(colonsLeft);
    colons->add(colonsRight);
    result = new ASTBaseType("bool");
    
  } else {
    panic("Type inference for operator '%s' not implemented", op);
  }
    
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTPrintExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);
  
  ASTType *result = followExp->getType(ctx, colons);
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTLetBindingExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  /* Type the right-hand expression first, and remember these colons separately */

  TCColons *colonsRightHand = new TCColons();
  ASTType *rightHandType = rightHandExp->getType(ctx, colonsRightHand);
  if (!rightHandType) 
    return NULL;

  /* If a type was asserted for the following expression, it better have that type */
  
  if (assertedTypeOrNull && !rightHandType->equals(assertedTypeOrNull, ctx)) {
    ctx->addError(this, "Type of expression bound to '%s' was asserted to be '%s', but was inferred as '%s'", nameToBind, render(assertedTypeOrNull), render(rightHandType));
    return NULL;
  }
  
  /* Now bind the symbol and type the following expression */

  ctx->addSymbol(nameToBind, rightHandType->clone());

  ASTType *resultType = followExp->getType(ctx, colons);
  if (!resultType) 
    return NULL;
  
  /* Update the colons */
  
  double colon = colons->containsSymbol(nameToBind) ? colons->getSymbol(nameToBind) : 0;
  colons->removeSymbol(nameToBind);
  colonsRightHand->multiplyBy(colon);
  colons->add(colonsRightHand);
  
  END_TYPE(this, resultType, ctx, colons);
  return resultType;
}

ASTType *ASTApplyExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  TCColons *argColons = new TCColons();
  int argTypeHandle = ctx->storeType(arg->getType(ctx, argColons));
  ASTType *funTypeNamed = fun->getType(ctx, colons);
  ASTType *argType = ctx->retrieveType(argTypeHandle);
  
  if (!funTypeNamed || !argType) 
    return NULL;
  
  ASTType *funType = funTypeNamed->cloneAndResolveIfNamed(ctx);
  
  if (funType->getTypeCode() != TYP_LAMBDA) {
    printf("Applying %s to %s\n", render(funType), render(argType));
    printf("... in %s\n", render(this));
    panic("Don't know how to type application of a non-lambda expression");
  }

  ASTFunctionType *funTypeAsFunctionType = (ASTFunctionType*) funType;

  ASTType *funTypeArg = funTypeAsFunctionType->getArgType();
  if (!ASTType::unify(&argType, &funTypeArg, ctx, &funType)) {
    ctx->addError(this, "Trying to apply a '%s' function to a '%s'", render(funType), render(argType));
    return NULL;
  }

  argColons->multiplyBy(funTypeAsFunctionType->getSensitivity());
  colons->add(argColons);
    
  ASTType *result = ((ASTFunctionType*)funType)->getResultType()->clone();
    
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTTensorTuple::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  int rightTypeHandle = ctx->storeType(right->getType(ctx, colons));
  ASTType *leftType = left->getType(ctx, colons);
  ASTType *rightType = ctx->retrieveType(rightTypeHandle);
  if (!rightType || !leftType)
    return NULL;
  
  ASTType *result = new ASTTensorType(leftType, rightType);
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTAmpersandTuple::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  TCColons *leftColons = new TCColons();
  int leftTypeHandle = ctx->storeType(left->getType(ctx, leftColons));

  TCColons *rightColons = new TCColons();
  ASTType *rightType = right->getType(ctx, rightColons);
  ASTType *leftType = ctx->retrieveType(leftTypeHandle);

  if (!leftType || !rightType) 
    return NULL;

  if (leftColons->containsNonzero()) {
    ctx->addError(this, "Left element of ampersand tuple contains colons");
    return NULL;
  }
  
  if (rightColons->containsNonzero()) {
    ctx->addError(this, "Right element of ampersand tuple contains colons");
    return NULL;
  }

  ASTType *result = new ASTAmpersandType(leftType, rightType);
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTTensorDestructorExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  /* Type the expression that produces the tensor, and keep its colons on the side */

  TCColons *producingColons = new TCColons();
  ASTType *producingTypeNamed = producingExp->getType(ctx, producingColons);
  if (!producingTypeNamed) 
    return NULL;
  
  ASTType *producingType = producingTypeNamed->cloneAndResolveIfNamed(ctx);
  if (producingType->getTypeCode() != TYP_TENSOR) {
    ctx->addError(this, "Producing expression does not return a tensor type");
    return NULL;
  }
  
  /* Now type the expression that consumes the tensor elements */
  
  ASTTensorType *tty = (ASTTensorType*)producingType;
  ctx->addSymbol(name1, tty->getLeft()->clone());
  ctx->addSymbol(name2, tty->getRight()->clone());
  
  ASTType *consumingType = consumingExp->getType(ctx, colons);
  if (!consumingType)
    return NULL;
  
  /* Update the colons */
  
  double colon1 = colons->containsSymbol(name1) ? colons->getSymbol(name1) : 0;
  double colon2 = colons->containsSymbol(name2) ? colons->getSymbol(name2) : 0;

  ctx->removeSymbol(name1);
  ctx->removeSymbol(name2);  
  colons->removeSymbol(name1);
  colons->removeSymbol(name2);

  double maxcolon = (colon1>colon2) ? colon1 : colon2;
  producingColons->multiplyBy(maxcolon);
  colons->add(producingColons);
  
  ASTType *result = consumingType;

  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTInjectExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  ASTType *innerType = innerExp->getType(ctx, colons);
  if (!innerType)
    return NULL;
  
  ASTType *typeArgResolved = typeArg->cloneAndResolveIfNamed(ctx);
  if (typeArgResolved->getTypeCode() != TYP_UNION) {
    ctx->addError(this, "Type argument of inl/r is not a union type");
    return NULL;
  }

  ASTUnionType *typeAsUnion = (ASTUnionType*) typeArgResolved;
  if (!innerType->equals(typeAsUnion->getOption(branch), ctx)) {
    ctx->addError(this, "Argument of in%s has type %s, but we need %s", branch ? "r" : "l", render(innerType), render(typeAsUnion->getOption(branch)));
    return NULL;
  }
    
  ASTType *result = typeArg->clone();
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTCaseExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  TCColons *exprColons = new TCColons();
  ASTType *exprTypeNamed = expr->getType(ctx, exprColons);
  if (!exprTypeNamed)
    return NULL;
  
  ASTType *exprType = exprTypeNamed->cloneAndResolveIfNamed(ctx);
  if (exprType->getTypeCode() != TYP_UNION) {
    ctx->addError(this, "Case argument does not return a union type");
    return NULL;
  }
  
  ASTUnionType *tun = (ASTUnionType*)exprType;
  int tunHandle = ctx->storeType(tun);
  
  TCColons *opt0colons = new TCColons();
  ctx->addSymbol(id[0], tun->getOption(0)->clone());
  int type0handle = ctx->storeType(branch[0]->getType(ctx, opt0colons));
  ctx->removeSymbol(id[0]);
  double colon0 = opt0colons->containsSymbol(id[0]) ? opt0colons->getSymbol(id[0]) : 0;
  opt0colons->removeSymbol(id[0]);

  tun = (ASTUnionType*) ctx->retrieveType(tunHandle);
  TCColons *opt1colons = new TCColons();
  ctx->addSymbol(id[1], tun->getOption(1)->clone());
  ASTType *type1 = branch[1]->getType(ctx, opt1colons);
  ctx->removeSymbol(id[1]);
  double colon1 = opt1colons->containsSymbol(id[1]) ? opt1colons->getSymbol(id[1]) : 0;
  opt1colons->removeSymbol(id[1]);
  
  ASTType *type0 = ctx->retrieveType(type0handle);
  
  if (!type0 || !type1)
    return NULL;
  
  if (!type0->equals(type1, ctx)) {
    ctx->addError(this, "Case branches return different types");
    return NULL;
  }

  double maxColon = (colon0 > colon1) ? colon0 : colon1;
  opt0colons->takeMax(opt1colons);
  exprColons->multiplyBy(maxColon);
  colons->add(exprColons);
  colons->add(opt0colons);

  ASTType *result = type0;
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTIfThenElseExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  TCColons *exprColons = new TCColons();
  ASTType *exprTypeNamed = expr->getType(ctx, exprColons);
  if (!exprTypeNamed)
    return NULL;
  
  ASTType *exprTypePrelim = exprTypeNamed->cloneAndResolveIfNamed(ctx);
  if (exprTypePrelim->getTypeCode() != TYP_BASE) {
    ctx->addError(this, "If argument does not return a base type");
    return NULL;
  }
  
  ASTBaseType *exprTypeAsBase = (ASTBaseType*) exprTypePrelim;
  if (strcmp(exprTypeAsBase->getName(), "bool")) {
    ctx->addError(this, "If argument does not return a bool");
    return NULL;
  }
  
  int exprHandle = ctx->storeType(exprTypePrelim);
  
  TCColons *thenColons = new TCColons();
  int thenHandle = ctx->storeType(thenBranch->getType(ctx, thenColons));

  TCColons *elseColons = new TCColons();
  ASTType *elseType = elseBranch->getType(ctx, elseColons);
  ASTType *thenType = ctx->retrieveType(thenHandle);
  ASTType *exprType = ctx->retrieveType(exprHandle);

  if (!thenType || !elseType)
    return NULL;
  
  if (!thenType->equals(elseType, ctx)) {
    ctx->addError(this, "Then and else branch return different types");
    return NULL;
  }

  colons->add(exprColons);
  thenColons->takeMax(elseColons);
  colons->add(thenColons);

  ASTType *result = thenType;
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTProjectExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  ASTType *innerTypeNamed = innerExp->getType(ctx, colons);
  if (!innerTypeNamed) 
    return NULL;
  
  ASTType *innerType = innerTypeNamed->cloneAndResolveIfNamed(ctx);
  if (innerType->getTypeCode() != TYP_AMPERSAND) {
    ctx->addError(this, "Producing expression does not return an ampersand type");
    return NULL;
  }
  
  ASTAmpersandType *amp = (ASTAmpersandType*) innerType;
  ASTType *result = branch ? amp->getRight()->clone() : amp->getLeft()->clone();

  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTFoldExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  ASTType *typeArgResolved = typeArg->cloneAndResolveIfNamed(ctx);
  if (typeArgResolved->getTypeCode() != TYP_MU) {
    ctx->addError(this, "Cannot type fold because type argument is not a mu type");
    return NULL;
  }
  
  ASTMuType *muTypeArg = (ASTMuType*) typeArgResolved;
  ASTType *unrolledType = muTypeArg->unroll();

  ASTType *innerType = innerExp->getType(ctx, colons);
  if (!innerType) 
    return NULL;

  if (!innerType->equals(unrolledType, ctx)) {
    ctx->addError(this, "Inner type of fold is '%s', but expected unrolled type '%s'", render(innerType), render(unrolledType));
    return NULL;
  }
  
  ASTType *result = typeArg->clone();
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTUnfoldExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  ASTType *innerTypeNamed = innerExp->getType(ctx, colons);
  if (!innerTypeNamed)
    return NULL;

  ASTType *innerType = innerTypeNamed->cloneAndResolveIfNamed(ctx);
  if (innerType->getTypeCode() != TYP_MU) {
    ctx->addError(this, "Cannot type unfold because inner type is not a mu type");
    return NULL;
  }
  
  ASTType *result = ((ASTMuType*)innerType)->unroll();
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTTypeDefExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);

  ctx->addSymbol(nameToBind, typeDef->clone());
  ASTType *result = followExp->getType(ctx, colons);
  if (!result)
    return NULL;
  
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTSampleExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);
  
  ASTType *rightHandTypeNamed = rightHandExp->getType(ctx, colons);
  if (!rightHandTypeNamed)
    return NULL;
  
  ASTType *rightHandType = rightHandTypeNamed->cloneAndResolveIfNamed(ctx);
  if (rightHandType->getTypeCode() != TYP_FUZZY) {
    ctx->addError(this, "Cannot type sample because right hand type is not a fuzzy type");
    return NULL;
  }

  ctx->addSymbol(nameToBind, ((ASTFuzzyType*)rightHandType)->getInnerType()->clone());
  assert(!colons->containsSymbol(nameToBind));
  colons->addSymbol(nameToBind, INFINITY);

  ASTType *result = followExp->getType(ctx, colons);
  if (!result) 
    return NULL;
  
  ctx->removeSymbol(nameToBind);
    
  END_TYPE(this, result, ctx, colons);
  return result;
}

ASTType *ASTSetExpression::getType(TCContext *ctx, TCColons *colons)
{
  BEGIN_TYPE(this, ctx, colons);
  
  ASTType *result = new ASTSetType(memberType->clone(), isMulti);
  
  END_TYPE(this, result, ctx, colons);
  return result;
}
